from zvartools.candidate import VariabilityCandidate
from zvartools.external import (
    connect_to_kowalski,
    query_gaia,
    query_ps1,
    query_2mass,
    query_allwise,
    query_galex,
)

credentials_path = "./credentials.json"

k = connect_to_kowalski(credentials_path)

candidate = VariabilityCandidate(
    psid=149053599892210341,
    ra=359.9892272949219,
    dec=34.20810317993164,
    valid=1.0,
    freq=None,
    fap=None,
    best_M=5,
)


def test_gaia_xmatch():
    xmatch_per_candidate = query_gaia(
        k, [candidate.id], [candidate.ra], [candidate.dec], 5.0
    )
    assert 149053599892210341 in xmatch_per_candidate

    xmatch = xmatch_per_candidate[149053599892210341]

    assert xmatch["id"] == 2875126085188101504
    assert xmatch["ra"] == 359.98922725671724
    assert xmatch["dec"] == 34.208105082089176
    assert xmatch["parallax"] == 4.6704942599886765
    assert xmatch["parallax_error"] == 0.039022523999999996
    assert xmatch["pmra"] == 17.838292582751816
    assert xmatch["pmra_error"] == 0.03712507
    assert xmatch["pmdec"] == 2.6669618109416886
    assert xmatch["pmdec_error"] == 0.026505275
    assert xmatch["astrometric_n_good_obs_al"] == 400
    assert xmatch["astrometric_chi2_al"] == 436.64474
    assert xmatch["ruwe"] == 1.0316383
    assert xmatch["phot_g_mean_flux"] == 8759.786131134488
    assert xmatch["phot_g_mean_flux_error"] == 6.2475442999999995
    assert xmatch["phot_g_mean_mag"] == 15.831133
    assert xmatch["phot_bp_mean_flux"] == 2002.6571921151572
    assert xmatch["phot_bp_mean_flux_error"] == 11.913191000000001
    assert xmatch["phot_bp_mean_mag"] == 17.084526
    assert xmatch["phot_rp_mean_flux"] == 10197.112668226617
    assert xmatch["phot_rp_mean_flux_error"] == 18.354717
    assert xmatch["phot_rp_mean_mag"] == 14.726703
    assert xmatch["phot_bp_rp_excess_factor"] == 1.3927017


def test_ps1_xmatch():
    xmatch_per_candidate = query_ps1(k, [candidate.id])
    assert 149053599892210341 in xmatch_per_candidate

    xmatch = xmatch_per_candidate[149053599892210341]

    assert xmatch["id"] == 149053599892210341 == candidate.id
    assert xmatch["gMeanPSFMag"] == 17.488899
    assert xmatch["gMeanPSFMagErr"] == 0.0063439999
    assert xmatch["rMeanPSFMag"] == 16.256399
    assert xmatch["rMeanPSFMagErr"] == 0.0049950001
    assert xmatch["iMeanPSFMag"] == 15.1836
    assert xmatch["iMeanPSFMagErr"] == 0.0035959998999999998
    assert xmatch["zMeanPSFMag"] == 14.690999999999999
    assert xmatch["zMeanPSFMagErr"] == 0.0051589999
    assert xmatch["yMeanPSFMag"] == 14.4642
    assert xmatch["yMeanPSFMagErr"] == 0.003332


def test_2mass_xmatch():
    xmatch_per_candidate = query_2mass(
        k, [candidate.id], [candidate.ra], [candidate.dec], 5.0
    )
    assert 149053599892210341 in xmatch_per_candidate

    xmatch = xmatch_per_candidate[149053599892210341]

    assert xmatch["id"] == "23595739+3412291"
    assert xmatch["ra"] == 359.989154
    assert xmatch["j_m"] == 13.286
    assert xmatch["j_cmsig"] == 0.019
    assert xmatch["h_m"] == 12.595
    assert xmatch["h_cmsig"] == 0.023
    assert xmatch["k_m"] == 12.387
    assert xmatch["k_cmsig"] == 0.022


def test_allwise_xmatch():
    xmatch_per_candidate = query_allwise(
        k, [candidate.id], [candidate.ra], [candidate.dec], 5.0
    )
    assert 149053599892210341 in xmatch_per_candidate

    xmatch = xmatch_per_candidate[149053599892210341]

    assert xmatch["id"] == "J235957.40+341229.0"
    assert xmatch["ra"] == 359.98920110000006
    assert xmatch["dec"] == 34.2080776
    assert xmatch["w1mpro"] == 12.257
    assert xmatch["w1sigmpro"] == 0.023
    assert xmatch["w2mpro"] == 12.175999999999998
    assert xmatch["w2sigmpro"] == 0.023
    assert xmatch["w3mpro"] == 11.832
    assert xmatch["w3sigmpro"] == 0.258
    assert xmatch["w4mpro"] == 8.606
    assert xmatch["w4sigmpro"] == 0.389


def test_galex_xmatch():
    xmatch_per_candidate = query_galex(
        k, [candidate.id], [candidate.ra], [candidate.dec], 5.0
    )
    assert 149053599892210341 in xmatch_per_candidate

    xmatch = xmatch_per_candidate[149053599892210341]

    assert xmatch["id"] == "GALEX J235957.3+341227"
    assert xmatch["ra"] == 359.989152
    assert xmatch["dec"] == 34.207653
    assert xmatch["b"] == 1
    assert xmatch["NUVmag"] == 22.1035
    assert xmatch["e_NUVmag"] == 0.3238



from typing import List, Tuple

import numpy as np


def remove_nondetections(
    time: np.ndarray, flux: np.ndarray, flux_err: np.ndarray, filter: np.ndarray
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Remove data points with flux = NaN

    Parameters
    ----------
    time : np.ndarray
        Time of observation
    flux : np.ndarray
        Flux of the object
    flux_err : np.ndarray
        Error in the flux
    filter : np.ndarray
        Filter used for the observation

    Returns
    -------
    Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]
        Time, flux, flux_err, filter with NaN values removed
    """
    mask = np.isnan(flux)
    return (
        time[~mask],
        flux[~mask],
        flux_err[~mask],
        filter[~mask],
    )


def get_cdf(data: np.ndarray) -> np.ndarray:
    """
    Get the Cumulative Distribution Function (CDF) of the data

    Parameters
    ----------
    data : np.ndarray
        Data to calculate the CDF for

    Returns
    -------
    np.ndarray
        Bin centers and CDF values
    """
    # Filter data to include only values greater than 0
    filtered_data = data[data > 0]

    # Create a distribution of the filtered data
    hist, bin_edges = np.histogram(filtered_data, bins=100, density=True)
    bin_centers = (bin_edges[:-1] + bin_edges[1:]) / 2

    # Find the CDF of the distribution
    cdf = np.cumsum(hist) * np.diff(bin_edges)
    return bin_centers, cdf


def get_extrema(data: np.ndarray, cdf_value: float = 0.999) -> np.ndarray:
    """
    Get the extrema of the data

    Parameters
    ----------
    data : np.ndarray
        Data to calculate the extrema for (e.g. flux)
    cdf_value : float, optional
        CDF value to use as the threshold, by default 0.999

    Returns
    -------
    np.ndarray
        Indices of the extrema
    """
    bin_centers, cdf = get_cdf(data)

    threshold_index = np.searchsorted(cdf, cdf_value)
    threshold = bin_centers[threshold_index]

    candidate_indices = data > threshold
    return np.array(candidate_indices)


def get_string_length(
    candidate: object,
    photometry: List[np.ndarray],
    period: float = None,
) -> float:
    """
    Calculate the string length of the candidate

    Parameters
    ----------
    candidate : VariableCandidate
        Variable candidate
    photometry : List[np.ndarray]
        Photometry data
    period : float, optional
        Period of the candidate, by default None

    Returns
    -------
    float
        String length of the candidate
    """
    if period is None:
        period = 1 / candidate.freq
    # remove data points with flux = NaN
    time, flux, _, _ = remove_nondetections(
        photometry[0], photometry[1], photometry[2], photometry[3]
    )

    phase = (time / (period * 86400)) % 2  # period is converted from days to seconds

    sorted_indices = np.argsort(phase)
    phase = phase[sorted_indices]
    flux = flux[sorted_indices]

    # Calculate string length as sum of distances between consecutive points
    string_length = 0.0
    for i in range(len(phase) - 1):
        delta_phase = phase[i + 1] - phase[i]
        delta_flux = flux[i + 1] - flux[i]
        string_length += np.sqrt(delta_phase**2 + delta_flux**2)

    # Optionally: Close the loop by connecting the last point back to the first
    delta_phase = 1.0 - phase[-1] + phase[0]  # wrap-around in phase
    delta_flux = flux[-1] - flux[0]
    string_length += np.sqrt(delta_phase**2 + delta_flux**2)

    string_length = string_length / len(phase)

    return string_length


def get_entropy(
    candidate: object,
    photometry: List[np.ndarray],
    period: float = None,
    num_bins: int = 10,
):
    """
    Calculate the entropy of the candidate

    Parameters
    ----------
    candidate : VariableCandidate
        Variable candidate
    photometry : List[np.ndarray]
        Photometry data
    period : float, optional
        Period of the candidate, by default None
    num_bins : int, optional
        Number of bins to use for the entropy calculation, by default 10

    Returns
    -------
    Tuple[float, float, float, float]
        Entropy, minimum entropy, negative entropy expectation, sigma expectation
    """
    if period is None:
        period = 1 / candidate.freq
    # remove data points with flux = NaN
    time, flux, _, _ = remove_nondetections(
        photometry[0], photometry[1], photometry[2], photometry[3]
    )

    phase = (time / (period * 86400)) % 2  # period is converted from days to seconds

    # fold the period (flux) to have the points in the same order as the phase
    sorted_indices = np.argsort(phase)
    phase = phase[sorted_indices]
    flux = flux[sorted_indices]

    num_in_bin = np.histogram(flux, num_bins)[0]
    ntot = np.sum(num_in_bin)
    # print("Num in bin,ntot: ",num_in_bin,ntot)
    frac_bin = num_in_bin / ntot + 10 ** (
        -10
    )  # Note: 10**(-10) takes care of zero bin values

    entropy = np.sum((frac_bin) * np.log2(frac_bin))

    min_entropy = np.log2(1 / num_bins)

    beta = 1.0  # dummy value
    binsx = num_bins
    sigma_expect = (
        beta * (1 / ntot) * np.sqrt((binsx - 1) / 2 + (binsx**2 - 1) / (6 * ntot))
    )  # from Hutcheson
    entropy_expect = (
        np.log2(binsx)
        - (binsx - 1) / (2 * ntot)
        - (binsx - 1) * (binsx + 1) / (12 * ntot**2)
    )  # new version, uses Hutcheson thesis

    return entropy, min_entropy, -entropy_expect, sigma_expect  # Note minus sign



ALLOWED_BANDS = ["g", "r"]

FILTERS = ["g", "r"]
FILTER2IDX = {"g": 1, "r": 2}

BIN_IDX_TO_FREQ_COL = {0: "frequency_20", 1: "frequency_10", 2: "frequency_5"}
BIN_IDX_TO_FAP_COL = {0: "FAP_20", 1: "FAP_10", 2: "FAP_5"}



import json
import os
from typing import List
from typing import Tuple

import numpy as np
from penquins import Kowalski

from zvartools.spatial import great_circle_distance


def connect_to_kowalski(path_credentials: str, verbose: bool = True) -> Kowalski:
    """
    Connect to Kowalski using the credentials file

    Parameters
    ----------
    path_credentials : str
        Path to the JSON file with the Kowalski credentials
    verbose : bool
        Verbosity level

    Returns
    -------
    Kowalski
        Kowalski instance
    """
    if path_credentials is None:
        raise ValueError("Credentials file must be specified")
    if not os.path.exists(path_credentials):
        raise ValueError(f"Credentials file {path_credentials} does not exist")
    with open(path_credentials) as f:
        passwd = json.load(f)
    username = passwd["melman"]["username"]
    password = passwd["melman"]["password"]

    k = Kowalski(
        protocol="https",
        host="melman.caltech.edu",
        port=443,
        username=username,
        password=password,
        verbose=verbose,
        timeout=600,
    )
    return k


def validate_gaia_source(source: dict) -> Tuple[bool, str]:
    """
    Validate a Gaia source

    Parameters
    ----------
    source : dict

    Returns
    -------
    Tuple[bool, str]
        Validity of the source and the reason for invalidity
    """
    if not isinstance(source, dict):
        return False, "Source must be a dictionary"
    for band in ["g", "bp", "rp"]:
        mag = source.get(f"phot_{band}_mean_mag")
        flux = source.get(f"phot_{band}_mean_flux")
        flux_error = source.get(f"phot_{band}_mean_flux_error")
        if not all(isinstance(val, (int, float)) for val in [mag, flux, flux_error]):
            return False, f"{band} band magnitude, flux, and flux error must be numbers"
        if mag < 0:
            return False, f"{band} band magnitude must be positive"
        if np.abs(flux) / flux_error < 5:
            return False, f"{band} band SNR must be greater than 5"

    ra, dec = source.get("ra"), source.get("dec")
    if not all(isinstance(val, (int, float)) for val in [ra, dec]):
        return False, "RA and Dec must be numbers"

    return True, None


def validate_positional_source(source: dict) -> Tuple[bool, str]:
    """
    Validate a positional source

    Parameters
    ----------
    source : dict

    Returns
    -------
    Tuple[bool, str]
        Validity of the source and the reason for invalidity
    """
    if not isinstance(source, dict):
        return False, "Source must be a dictionary"
    ra, dec = source.get("ra"), source.get("dec")
    if not all(isinstance(val, (int, float)) for val in [ra, dec]):
        return False, "RA and Dec must be numbers"
    return True, None


def query_cone_search(
    k: Kowalski,
    ids: List[int],
    ras: List[int],
    decs: List[int],
    radius: float,
    id_field: str = "_id",
    batch_size: int = 1000,
    catalog: str = None,
    projection: dict = None,
    validate: callable = None,
) -> dict:
    """
    Query Kowalski with a cone search

    Parameters
    ----------
    k : Kowalski
        Kowalski instance
    ids : List[int]
        List of IDs
    ras : List[int]
        List of RAs
    decs : List[int]
        List of Decs
    radius : float
        Radius in arcseconds
    id_field : str
        ID field
    batch_size : int
        Batch size
    catalog : str
        Catalog name
    projection : dict
        Projection
    validate : callable
        Validation function

    Returns
    -------
    dict
        Query results
    """
    if not isinstance(k, Kowalski):
        raise ValueError("Kowalski must be an instance of the Kowalski class")
    if not isinstance(ids, list):
        raise ValueError("IDs must be provided as a list")
    if not isinstance(ras, list):
        raise ValueError("RAs must be provided as a list")
    if not isinstance(decs, list):
        raise ValueError("Decs must be provided as a list")
    if not isinstance(radius, (int, float)):
        raise ValueError("Radius must be a number")
    if not isinstance(batch_size, int):
        raise ValueError("Batch size must be an integer")
    if not isinstance(catalog, str):
        raise ValueError("Catalog must be a string")
    if not isinstance(projection, dict):
        raise ValueError("Projection must be a dictionary")
    if not callable(validate) and validate is not None:
        raise ValueError("Validate must be a callable, or None")

    # we want a mapper from id to (ra, dec)
    # so that when we loop over the xmatch results, we can easily find the original source
    # position and compute angular separation to each xmatch result
    id_to_position = {id: (ra, dec) for id, ra, dec in zip(ids, ras, decs)}

    # Create a list of tuples of (psid, ra, dec) for each source
    inputs = []
    for i in range(len(ids)):
        inputs.append((ids[i], ras[i], decs[i]))

    # Create batches of sources
    batches = [inputs[i : i + batch_size] for i in range(0, len(inputs), batch_size)]

    queries = []
    for batch in batches:
        # Create a list of dictionaries for each source
        query = {
            "query_type": "cone_search",
            "query": {
                "object_coordinates": {
                    "cone_search_radius": radius,
                    "cone_search_unit": "arcsec",
                    "radec": {
                        str(id): [float(ra), float(dec)] for id, ra, dec in batch
                    },
                },
                "catalogs": {
                    catalog: {
                        "filter": {},
                        "projection": projection,
                    }
                },
            },
        }
        queries.append(query)

    responses = k.query(queries=queries, use_batch_query=True, max_n_threads=10)
    responses = responses.get("default")

    results = {}
    for response in responses:
        if response.get("status") != "success":
            print(f"Query failed with error: {response.get('message')}")
            continue
        for id, result in response.get("data").get(catalog).items():
            # Filter out sources that don't pass the quality cuts
            result_filtered = []
            if validate is None:
                result_filtered = result
            else:
                for source in result:
                    valid, _ = validate(source)
                    if valid:
                        result_filtered.append(source)
                    else:
                        # print(f"Source {id} failed quality cuts: {reason}")
                        pass
            result = result_filtered

            for source in result:
                try:
                    source["id"] = int(source[id_field])
                except ValueError:
                    source["id"] = str(source[id_field]).strip()
                source.pop(id_field)

            if len(result) == 0:
                results[int(id)] = None
                continue
            if len(result) == 1:
                results[int(id)] = result[0]
                continue

            # keep the closest source
            ra, dec = id_to_position[int(id)]
            min_dist = np.inf
            for source in result:
                dist = great_circle_distance(ra, dec, source["ra"], source["dec"])
                if dist < min_dist:
                    min_dist = dist
                    results[int(id)] = source

    return results


def query_by_id(
    k: Kowalski,
    ids: List[int],
    id_field: str = "_id",
    batch_size: int = 1000,
    catalog: str = None,
    projection: dict = None,
):
    """
    Query Kowalski by ID

    Parameters
    ----------
    k : Kowalski
        Kowalski instance
    ids : List[int]
        List of IDs
    id_field : str
        ID field
    batch_size : int
        Batch size
    catalog : str
        Catalog name
    projection : dict
        Projection

    Returns
    -------
    dict
        Query results
    """
    if not isinstance(k, Kowalski):
        raise ValueError("Kowalski must be an instance of the Kowalski class")
    if not isinstance(ids, list):
        raise ValueError("IDs must be provided as a list")
    if not isinstance(id_field, str):
        raise ValueError("ID field must be a string")
    if not isinstance(batch_size, int):
        raise ValueError("Batch size must be an integer")
    if not isinstance(catalog, str):
        raise ValueError("Catalog must be a string")
    if not isinstance(projection, dict):
        raise ValueError("Projection must be a dictionary")

    # Create batches of sources
    batches = [ids[i : i + batch_size] for i in range(0, len(ids), batch_size)]

    # if the type of the id is uint64, convert it to int
    for i, batch in enumerate(batches):
        batches[i] = [int(id) if isinstance(id, np.uint64) else id for id in batch]

    queries = []
    for batch in batches:
        # Create a list of dictionaries for each source
        query = {
            "query_type": "find",
            "query": {
                "catalog": catalog,
                "filter": {"_id": {"$in": batch}},
                "projection": projection,
            },
        }
        queries.append(query)

    responses = k.query(queries=queries, use_batch_query=True, max_n_threads=10)
    responses = responses.get("default")

    results = {}
    for response in responses:
        if response.get("status") != "success":
            print(f"Query failed with error: {response.get('message')}")
            continue
        for result in response.get("data"):
            id = result[id_field]
            result["id"] = result.pop(id_field)
            results[id] = result

    # for those that didn't return a result, add None
    for id in ids:
        if id not in results:
            results[id] = None

    return results


def query_ps1(
    k: Kowalski,
    ids: List[int],
    batch_size: int = 1000,
):
    """
    Query PanSTARRS1

    Parameters
    ----------
    k : Kowalski
        Kowalski instance
    ids : List[int]
        List of IDs
    batch_size : int
        Batch size

    Returns
    -------
    dict
        Query results
    """
    return query_by_id(
        k=k,
        ids=ids,
        id_field="_id",
        batch_size=batch_size,
        catalog="PS1_DR1",
        projection={
            "_id": 1,
            "ra": 1,
            "dec": 1,
            "gMeanPSFMag": 1,
            "gMeanPSFMagErr": 1,
            "rMeanPSFMag": 1,
            "rMeanPSFMagErr": 1,
            "iMeanPSFMag": 1,
            "iMeanPSFMagErr": 1,
            "zMeanPSFMag": 1,
            "zMeanPSFMagErr": 1,
            "yMeanPSFMag": 1,
            "yMeanPSFMagErr": 1,
        },
    )


def query_gaia(
    k: Kowalski,
    ids: List[int],
    ras: List[int],
    decs: List[int],
    radius: float,
    batch_size: int = 1000,
):
    """
    Query Gaia

    Parameters
    ----------
    k : Kowalski
        Kowalski instance
    ids : List[int]
        List of IDs
    ras : List[int]
        List of RAs
    decs : List[int]
        List of Decs
    radius : float
        Radius in arcseconds
    batch_size : int
        Batch size

    Returns
    -------
    dict
        Query results
    """
    return query_cone_search(
        k=k,
        ids=ids,
        ras=ras,
        decs=decs,
        radius=radius,
        batch_size=batch_size,
        catalog="Gaia_EDR3",
        projection={
            "_id": 1,
            "pmra": 1,
            "pmra_error": 1,
            "pmdec": 1,
            "pmdec_error": 1,
            "parallax": 1,
            "parallax_error": 1,
            "phot_g_mean_mag": 1,
            "phot_bp_mean_mag": 1,
            "phot_rp_mean_mag": 1,
            "phot_g_mean_flux": 1,
            "phot_bp_mean_flux": 1,
            "phot_rp_mean_flux": 1,
            "phot_g_mean_flux_error": 1,
            "phot_bp_mean_flux_error": 1,
            "phot_rp_mean_flux_error": 1,
            "astrometric_chi2_al": 1,
            "astrometric_n_good_obs_al": 1,
            "phot_bp_rp_excess_factor": 1,
            "ruwe": 1,
            "ra": 1,
            "dec": 1,
        },
        validate=validate_gaia_source,
    )


def query_2mass(
    k: Kowalski,
    ids: List[int],
    ras: List[int],
    decs: List[int],
    radius: float,
    batch_size: int = 1000,
):
    """
    Query 2MASS

    Parameters
    ----------
    k : Kowalski
        Kowalski instance
    ids : List[int]
        List of IDs
    ras : List[int]
        List of RAs
    decs : List[int]
        List of Decs
    radius : float
        Radius in arcseconds
    batch_size : int
        Batch size

    Returns
    -------
    dict
        Query results
    """
    return query_cone_search(
        k=k,
        ids=ids,
        ras=ras,
        decs=decs,
        radius=radius,
        batch_size=batch_size,
        catalog="2MASS_PSC",
        projection={
            "_id": 1,
            "ra": 1,
            "dec": "$decl",
            "j_m": 1,
            "j_cmsig": 1,
            "h_m": 1,
            "h_cmsig": 1,
            "k_m": 1,
            "k_cmsig": 1,
        },
        validate=validate_positional_source,
    )


def query_allwise(
    k: Kowalski,
    ids: List[int],
    ras: List[int],
    decs: List[int],
    radius: float,
    batch_size: int = 1000,
):
    """
    Query AllWISE

    Parameters
    ----------
    k : Kowalski
        Kowalski instance
    ids : List[int]
        List of IDs
    ras : List[int]
        List of RAs
    decs : List[int]
        List of Decs
    radius : float
        Radius in arcseconds
    batch_size : int
        Batch size

    Returns
    -------
    dict
        Query results
    """
    return query_cone_search(
        k=k,
        ids=ids,
        ras=ras,
        decs=decs,
        id_field="designation",
        radius=radius,
        batch_size=batch_size,
        catalog="AllWISE",
        projection={
            "designation": 1,
            "ra": 1,
            "dec": 1,
            "w1mpro": 1,
            "w1sigmpro": 1,
            "w2mpro": 1,
            "w2sigmpro": 1,
            "w3mpro": 1,
            "w3sigmpro": 1,
            "w4mpro": 1,
            "w4sigmpro": 1,
        },
        validate=validate_positional_source,
    )


def query_galex(
    k: Kowalski,
    ids: List[int],
    ras: List[int],
    decs: List[int],
    radius: float,
    batch_size: int = 1000,
):
    """
    Query GALEX

    Parameters
    ----------
    k : Kowalski
        Kowalski instance
    ids : List[int]
        List of IDs
    ras : List[int]
        List of RAs
    decs : List[int]
        List of Decs
    radius : float
        Radius in arcseconds
    batch_size : int
        Batch size

    Returns
    -------
    dict
        Query results
    """
    return query_cone_search(
        k=k,
        ids=ids,
        ras=ras,
        decs=decs,
        id_field="name",
        radius=radius,
        batch_size=batch_size,
        catalog="GALEX",
        projection={
            "ra": 1,
            "dec": 1,
            "name": 1,
            "b": 1,
            "NUVmag": 1,
            "e_NUVmag": 1,
        },
        validate=validate_positional_source,
    )



import os
from typing import List, Union

import matplotlib.pyplot as plt
import numpy as np
from astropy.table import Table
from matplotlib.colors import LogNorm

from zvartools.candidate import VariabilityCandidate
from zvartools.enums import ALLOWED_BANDS
from zvartools.lightcurves import freq_grid, remove_nondetections

BAND_TO_COLOR = {1: "green", 2: "red", 3: "orange"}
BAND_IDX_TO_NAME = {1: "g", 2: "r", 3: "i"}
BAND_NAME_TO_IDX = {"g": 1, "r": 2, "i": 3}
MARKER_STYLES = {5: ("s", 20), 10: ("o", 20), 20: ("*", 35)}


def plot_gaia_cmd(
    candidate_list: List[VariabilityCandidate],
    figsize: tuple = (9, 8),
    output_path: str = None,
    show_plot: bool = True,
    ax=None,
    title_size: int = 16,
):
    """
    Plot a Gaia HR diagram with variability candidates

    Parameters
    ----------
    candidate_list : List[VariabilityCandidate]
        List of variability candidates
    figsize : tuple, optional
        Figure size, by default (9, 8)
    output_path : str, optional
        Path to save the plot, by default None
    show_plot : bool, optional
        Whether to show the plot, by default True
    ax : matplotlib.axes.Axes, optional
        Axes to plot on, by default None
    title_size : int, optional
        Font size for the title, by default 16

    Raises
    ------
    ValueError
        If candidates are not provided as a list
        If candidates are not of type Candidate
        If output path is not a string or PathLike object
        If show plot is not a boolean
        If no candidates found with valid BP-RP and MG values
    """
    if not isinstance(candidate_list, list):
        raise ValueError("Candidates must be provided as a list")
    if not all(
        isinstance(candidate, VariabilityCandidate) for candidate in candidate_list
    ):
        raise ValueError("Candidates must be of type Candidate")
    if output_path is not None and not isinstance(output_path, (str, os.PathLike)):
        raise ValueError("Output path must be a string or a PathLike object")
    if not isinstance(show_plot, bool):
        raise ValueError("Show plot must be a boolean")

    valid_candidates_idx = [
        i
        for i, candidate in enumerate(candidate_list)
        if (
            candidate.gaia.BP_RP is not None
            and candidate.gaia.MG is not None
            and not np.isnan(candidate.gaia.BP_RP)
            and not np.isnan(candidate.gaia.MG)
        )
    ]
    if len(valid_candidates_idx) == 0:
        raise ValueError("No candidates found with valid BP-RP and MG values")

    periods = [1 / candidate_list[i].freq for i in valid_candidates_idx]
    best_M = [candidate_list[i].best_M for i in valid_candidates_idx]
    bp_rp = [candidate_list[i].gaia.BP_RP for i in valid_candidates_idx]
    mg = [candidate_list[i].gaia.MG for i in valid_candidates_idx]

    # Load a Gaia HR diagram
    sample_path = os.path.join(
        os.path.dirname(__file__), "./data/hrd_query_edr3_200pc.fits"
    )
    gaia_sample = Table.read(sample_path).to_pandas()
    gaia_bprp = (
        gaia_sample.phot_bp_mean_mag.values - gaia_sample.phot_rp_mean_mag.values
    )
    gaia_gmag = gaia_sample.phot_g_mean_mag.values + 5.0 * np.log10(
        gaia_sample.parallax.values / 100
    )

    if ax is None:
        _, ax = plt.subplots(1, 1, figsize=figsize)

    ax.hist2d(
        gaia_bprp,
        gaia_gmag,
        cmap="gray",
        cmin=3,
        cmax=30,
        bins=(700, 380),
        alpha=0.8,
    )
    ax.hist2d(
        gaia_bprp,
        gaia_gmag,
        cmap="gray_r",
        cmin=30,
        cmax=1000,
        bins=(700, 380),
        alpha=0.8,
    )

    # Scatter plot for candidates colored by period with logarithmic scale
    for m_value in MARKER_STYLES:
        indices = [i for i, m in enumerate(best_M) if m == m_value]
        if not indices:
            continue
        marker, size = MARKER_STYLES[m_value]
        sc = ax.scatter(
            [bp_rp[i] for i in indices],
            [mg[i] for i in indices],
            c=[periods[i] for i in indices],
            cmap="plasma",
            s=size,
            edgecolor="none",
            norm=LogNorm(),
            marker=marker,
            label=f"best_M = {m_value}",
        )

    # Add colorbar
    cbar = plt.colorbar(sc, ax=ax)
    cbar.set_label("Period (log days)", fontsize=title_size)

    ax.set_title("Gaia HR Diagram with Variability Candidates", fontsize=title_size)
    ax.set_ylabel("$M_G$ (mag)", fontsize=title_size)
    ax.set_xlabel("$BP-RP$ (mag)", fontsize=title_size)
    ax.set_ylim(17.0, -2.0)
    ax.set_yticks([0, 5, 10, 15])
    ax.minorticks_on()
    ax.tick_params(which="both", direction="in", top=True, right=True, labelsize=14)
    ax.set_axisbelow(True)
    ax.grid(c="silver", ls=":", lw=1)
    ax.legend()
    plt.tight_layout()

    # Save the plot
    if output_path:
        output_dir = os.path.dirname(output_path)
        os.makedirs(output_dir, exist_ok=True)
        plt.savefig(output_path, dpi=300)

    if show_plot:
        plt.show()

    plt.close()


def plot_2mass_cmd(
    candidate_list: List[VariabilityCandidate],
    figsize: tuple = (9, 8),
    output_path: str = None,
    show_plot: bool = True,
    ax=None,
    title_size: int = 16,
):
    if not isinstance(candidate_list, list):
        raise ValueError("Candidates must be provided as a list")
    if not all(
        isinstance(candidate, VariabilityCandidate) for candidate in candidate_list
    ):
        raise ValueError("Candidates must be of type Candidate")
    if output_path is not None and not isinstance(output_path, (str, os.PathLike)):
        raise ValueError("Output path must be a string or a PathLike object")
    if not isinstance(show_plot, bool):
        raise ValueError("Show plot must be a boolean")

    valid_candidates_idx = [
        i
        for i, candidate in enumerate(candidate_list)
        if (
            candidate.twomass is not None
            and candidate.twomass.j is not None
            and candidate.twomass.k is not None
            and candidate.gaia is not None
            and candidate.gaia.parallax is not None
            and not np.isnan(candidate.twomass.j)
            and not np.isnan(candidate.twomass.k)
            and not np.isnan(candidate.gaia.parallax)
        )
    ]

    if len(valid_candidates_idx) == 0:
        raise ValueError("No candidates found with valid 2MASS and Gaia values")

    periods = [1 / candidate_list[i].freq for i in valid_candidates_idx]
    best_M = [candidate_list[i].best_M for i in valid_candidates_idx]
    j_k = [
        candidate_list[i].twomass.j - candidate_list[i].twomass.k
        for i in valid_candidates_idx
    ]
    j_abs = [
        # Use gaia parallax to calculate the absolute magnitude in the J band
        candidate_list[i].twomass.j
        - 5 * np.log10(1e3 / candidate_list[i].gaia.parallax)
        + 5
        for i in valid_candidates_idx
    ]

    if len(periods) == 0:
        raise ValueError("No candidates found with valid J-K values")

    if ax is None:
        _, ax = plt.subplots(1, 1, figsize=figsize)

    for m_value in MARKER_STYLES:
        indices = [i for i, m in enumerate(best_M) if m == m_value]
        if not indices:
            continue
        marker, size = MARKER_STYLES[m_value]
        sc = ax.scatter(
            [j_k[i] for i in indices],
            [j_abs[i] for i in indices],
            c=[periods[i] for i in indices],
            cmap="plasma",
            s=size,
            edgecolor="none",
            norm=LogNorm(),
            marker=marker,
            label=f"best_M = {m_value}",
        )

    cbar = plt.colorbar(sc, ax=ax)
    cbar.set_label("Period (log days)", fontsize=title_size)

    ax.set_title("2MASS Color-Color Diagram", fontsize=title_size)
    ax.set_ylabel("$M_J$ (mag)", fontsize=title_size)
    ax.set_xlabel("$J-K$ (mag)", fontsize=title_size)
    ax.minorticks_on()
    ax.tick_params(which="both", direction="in", top=True, right=True, labelsize=14)
    ax.set_axisbelow(True)
    ax.grid(c="silver", ls=":", lw=1)
    ax.legend()
    plt.tight_layout()

    # Save the plot
    if output_path:
        output_dir = os.path.dirname(output_path)
        os.makedirs(output_dir, exist_ok=True)
        plt.savefig(output_path, dpi=300)

    if show_plot:
        plt.show()

    plt.close()


def plot_2mass_color(
    candidate_list: List[VariabilityCandidate],
    figsize: tuple = (9, 8),
    output_path: str = None,
    show_plot: bool = True,
    ax=None,
    title_size: int = 16,
):
    if not isinstance(candidate_list, list):
        raise ValueError("Candidates must be provided as a list")
    if not all(
        isinstance(candidate, VariabilityCandidate) for candidate in candidate_list
    ):
        raise ValueError("Candidates must be of type Candidate")
    if output_path is not None and not isinstance(output_path, (str, os.PathLike)):
        raise ValueError("Output path must be a string or a PathLike object")
    if not isinstance(show_plot, bool):
        raise ValueError("Show plot must be a boolean")

    valid_candidates_idx = [
        i
        for i, candidate in enumerate(candidate_list)
        if (
            candidate.twomass is not None
            and candidate.twomass.j is not None
            and candidate.twomass.k is not None
            and not np.isnan(candidate.twomass.j)
            and not np.isnan(candidate.twomass.k)
        )
    ]

    if len(valid_candidates_idx) == 0:
        raise ValueError("No candidates found with valid J-K values")

    periods = [1 / candidate_list[i].freq for i in valid_candidates_idx]
    best_M = [candidate_list[i].best_M for i in valid_candidates_idx]
    j_k = [
        candidate_list[i].twomass.j - candidate_list[i].twomass.k
        for i in valid_candidates_idx
    ]
    j = [candidate_list[i].twomass.j for i in valid_candidates_idx]

    if len(periods) == 0:
        raise ValueError("No candidates found with valid J-K values")

    # Plot candidates
    if ax is None:
        _, ax = plt.subplots(1, 1, figsize=figsize)

    # Scatter plot for candidates colored by period with logarithmic scale
    for m_value in MARKER_STYLES:
        indices = [i for i, m in enumerate(best_M) if m == m_value]
        if not indices:
            continue
        marker, size = MARKER_STYLES[m_value]
        sc = ax.scatter(
            [j_k[i] for i in indices],
            [j[i] for i in indices],
            c=[periods[i] for i in indices],
            cmap="plasma",
            s=size,
            edgecolor="none",
            norm=LogNorm(),
            marker=marker,
            label=f"best_M = {m_value}",
        )

    # Add colorbar
    cbar = plt.colorbar(sc, ax=ax)
    cbar.set_label("Period (log days)", fontsize=title_size)

    # Set plot appearances
    ax.set_title("2MASS Color-Color Diagram", fontsize=title_size)
    ax.set_ylabel("$J$ (mag)", fontsize=title_size)
    ax.set_xlabel("$J-K$ (mag)", fontsize=title_size)
    ax.minorticks_on()
    ax.tick_params(which="both", direction="in", top=True, right=True, labelsize=14)
    ax.set_axisbelow(True)
    ax.grid(c="silver", ls=":", lw=1)
    ax.legend()
    plt.tight_layout()

    # Save the plot
    if output_path:
        output_dir = os.path.dirname(output_path)
        os.makedirs(output_dir, exist_ok=True)
        plt.savefig(output_path, dpi=300)

    if show_plot:
        plt.show()

    plt.close()


def plot_folded_lightcurve(
    candidate: VariabilityCandidate,
    photometry: List[np.ndarray],
    bands: List[str] = ["g", "r"],
    period: float = None,
    figsize: tuple = (9, 8),
    output_path: str = None,
    show_plot: bool = True,
    ax=None,
    marker_size: int = 4,
    title_size: int = 16,
    bins: Union[int, None] = None,
    bin_method: str = "mean",
):
    """
    Plot the folded lightcurve of a candidate

    Parameters
    ----------
    candidate : VariabilityCandidate
        Variability candidate
    photometry : List[np.ndarray]
        Photometry data
    bands : List[str], optional
        Bands to plot, by default ["g", "r"]
    period : float, optional
        Period of the candidate, by default None
    figsize : tuple, optional
        Figure size, by default (9, 8)
    output_path : str, optional
        Path to save the plot, by default None
    show_plot : bool, optional
        Whether to show the plot, by default True
    ax : matplotlib.axes.Axes, optional
        Axes to plot on, by default None
    marker_size : int, optional
        Marker size for the plot, by default 4
    title_size : int, optional
        Font size for the title, by default 16
    bins : Union[int, None], optional
        Number of bins to use for binning the data, by default None
    bin_method : str, optional
        Method to use for binning the data, by default "mean"

    Raises
    ------
    ValueError
        If an invalid band is specified
        If no data points found in the specified bands
        If no valid data points found to plot_folded_lightcurve
    """
    if period is None:
        period = 1 / candidate.freq
    bands = list({band.lower() for band in bands})
    if not all(band in ALLOWED_BANDS for band in bands):
        raise ValueError(f"Invalid band specified. Allowed bands are {ALLOWED_BANDS}")
    bands = [BAND_NAME_TO_IDX[band] for band in bands]

    # remove data points with flux = NaN
    time, flux, fluxerr, filters = remove_nondetections(
        photometry[0], photometry[1], photometry[2], photometry[3]
    )

    # only keep data points in the band of interest, so where filters is in the list of bands
    mask = np.array([f in bands for f in filters])

    try:
        time = time[mask]
        flux = flux[mask]
        fluxerr = fluxerr[mask]
        filters = filters[mask]
    except IndexError:
        raise ValueError(
            f"No data points found in the specified bands for {int(candidate.psid)}"
        )

    if len(time) == 0:
        raise ValueError(
            f"No valid data points found to plot_folded_lightcurve for {int(candidate.psid)}"
        )
        # from bands, remove bands for which we have no data points
    bands = [band for band in bands if np.any(filters == band)]

    phase = (time / (period * 86400)) % 2  # period is converted from days to seconds

    if ax is None:
        _, ax = plt.subplots(1, 1, figsize=figsize)

    if bins is not None:
        # Create bins and calculate which bin each phase belongs to
        bin_indices = np.digitize(phase, np.linspace(0, 2, bins + 1)) - 1

        # we bin the points per band
        binned_flux = np.zeros((len(bands), bins))
        binned_fluxerr = np.zeros((len(bands), bins))
        for i, band in enumerate(bands):
            for j in range(bins):
                mask = (filters == band) & (bin_indices == j)
                # if there are no points in the bin, we set the flux to -inf
                if not np.any(mask):
                    binned_flux[i, j] = -np.inf
                    binned_fluxerr[i, j] = -np.inf
                    continue
                if bin_method == "mean":
                    binned_flux[i, j] = np.mean(flux[mask])
                    binned_fluxerr[i, j] = np.sqrt(np.sum(fluxerr[mask] ** 2)) / np.sum(
                        mask
                    )
                elif bin_method == "median":
                    binned_flux[i, j] = np.median(flux[mask])
                    binned_fluxerr[i, j] = np.median(fluxerr[mask])
                else:
                    raise ValueError(
                        "Invalid bin_method. Must be either 'mean' or 'median'"
                    )

        # # remove bins with no data
        mask = np.all(binned_flux != -np.inf, axis=0)
        binned_flux = binned_flux[:, mask]
        binned_fluxerr = binned_fluxerr[:, mask]
        x = np.arange(bins)
        x = x[mask]
        x = x / bins

        for i, band in enumerate(bands):
            ax.errorbar(
                x,
                binned_flux[i],
                yerr=binned_fluxerr[i],
                fmt="o",
                label=BAND_IDX_TO_NAME[band],
                color=BAND_TO_COLOR[band],
                ms=marker_size,
            )
    else:
        for band in bands:
            mask = filters == band
            if not np.any(mask):
                continue
            ax.errorbar(
                phase[mask] / 2,
                flux[mask],
                yerr=fluxerr[mask],
                fmt="o",
                label=band,
                color=BAND_TO_COLOR[band],
                ms=marker_size,
            )

    ax.set_xlabel("Phase")
    ax.set_ylabel("Flux")
    ax.set_title(
        f"ID: {int(candidate.psid)}\nRA: {candidate.ra:.4f}, DEC: {candidate.dec:.4f}\nPeriod: {24 / candidate.freq:.4f} hours, Best M: {candidate.best_M}",
        fontsize=title_size,
    )

    if output_path:
        output_dir = os.path.dirname(output_path)
        os.makedirs(output_dir, exist_ok=True)
        plt.savefig(output_path, dpi=300)

    if show_plot:
        plt.show()
        plt.close()


def plot_periodicity(
    candidate: VariabilityCandidate,
    photometry: List[np.ndarray],
    pgram: np.ndarray,
    period: float,
    period_unit: str = "hours",
    show_plot: bool = True,
    figsize: tuple = (12, 14),
    bins: Union[int, None] = None,
    bin_method: str = "mean",
):
    """
    Plot the lightcurve, periodogram, and phased lightcurve of a candidate

    Parameters
    ----------
    candidate : VariabilityCandidate
        Variability candidate
    photometry : List[np.ndarray]
        Photometry data
    pgram : np.ndarray
        Periodogram data
    period : float
        Period of the candidate, in hours
    period_unit : str, optional
        Unit of the period if provided, by default "hours"
    show_plot : bool, optional
        Whether to show the plot, by default True
    figsize : tuple, optional
        Figure size, by default (12, 14)
    bins : Union[int, None], optional
        Number of bins to use for binning the data, by default None
    bin_method : str, optional
        Method to use for binning the data, by default "mean"

    Raises
    ------
    ValueError
        If there are multiple filters
        If no valid data points found to plot_periodicity
    """
    if period is None:
        period = 1 / candidate.freq
    elif period_unit == "minutes":
        # convert from minutes to seconds
        period *= 60
    elif period_unit == "hours":
        # convert from hours to seconds
        period *= 60 * 60
    elif period_unit == "days":
        # convert from days to seconds
        period *= 86400
    # for now this method supports single band data, so to prevent any errors we throw an exception if there are multiple filters
    if len(set(photometry[3])) > 1:
        raise ValueError("This method only supports single band data (for now)")

    # remove data points with flux = NaN
    time, flux, fluxerr, _ = remove_nondetections(
        photometry[0], photometry[1], photometry[2], photometry[3]
    )
    if len(time) == 0:
        raise ValueError("No valid data points found to plot_periodicity")

    fgrid = freq_grid(time)

    phase = (time / period) % 2

    # we plot all 3 plots above in a row of 3 subplots
    fig, axs = plt.subplots(3, 1, figsize=figsize)

    # plot the lightcurve
    axs[0].errorbar(time, flux, yerr=fluxerr, fmt="o")
    axs[0].set_title("Lightcurve")
    axs[0].set_xlabel("Time (MJD)")

    # plot the periodogram
    axs[1].plot(fgrid, pgram)
    axs[1].set_xscale("log")
    axs[1].set_title("Periodogram")
    axs[1].set_xlabel("Frequency (Hz)")

    if bins is not None:
        # Create bins and calculate which bin each phase belongs to
        bin_indices = np.digitize(phase, np.linspace(0, 2, bins + 1)) - 1

        # we bin the points
        binned_flux = np.zeros(bins)
        binned_fluxerr = np.zeros(bins)
        for j in range(bins):
            mask = bin_indices == j
            # if there are no points in the bin, we set the flux to -inf
            if not np.any(mask):
                binned_flux[j] = -np.inf
                binned_fluxerr[j] = -np.inf
                continue
            if bin_method == "mean":
                binned_flux[j] = np.mean(flux[mask])
                binned_fluxerr[j] = np.sqrt(np.sum(fluxerr[mask] ** 2)) / np.sum(mask)
            elif bin_method == "median":
                binned_flux[j] = np.median(flux[mask])
                binned_fluxerr[j] = np.median(fluxerr[mask])
            else:
                raise ValueError(
                    "Invalid bin_method. Must be either 'mean' or 'median'"
                )

        # # remove bins with no data
        mask = binned_flux != -np.inf
        flux = binned_flux[mask]
        fluxerr = binned_fluxerr[mask]
        x = np.arange(bins)
        x = x[mask]
        # go back to phase
        x = x / bins
    else:
        x = phase / 2

    # plot the phased lightcurve
    axs[2].errorbar(x, flux, yerr=fluxerr, fmt="o")
    axs[2].set_title(f"Phased Lightcurve (Period: {period / 60 / 60} hours)")
    axs[2].set_xlabel("Phase")

    # add a title to the entire figure
    fig.suptitle(
        f"ID: {int(candidate.psid)}, RA: {candidate.ra}, DEC: {candidate.dec}\nBest M: {candidate.best_M}",
        fontsize=16,
    )

    # add spacing between the subplots
    plt.tight_layout()

    if show_plot:
        plt.show()


def plot_folded_photometry_stat_per_bin(
    candidate: VariabilityCandidate,
    photometry: List[np.ndarray],
    period: float = None,
    num_bins: int = 20,
    method: str = "median",
    figsize: tuple = (9, 8),
    output_path: str = None,
    show_plot: bool = True,
    ax=None,
    line_width: int = 1,
    title_size: int = 16,
):
    """
    Plot the median or mean flux per bin for the folded lightcurve

    Parameters
    ----------
    candidate : VariabilityCandidate
        Variability candidate
    photometry : List[np.ndarray]
        Photometry data
    period : float, optional
        Period of the candidate, by default None
    num_bins : int, optional
        Number of bins to use for binning the data, by default 20
    method : str, optional
        Method to use for calculating the flux per bin, by default "median"
    figsize : tuple, optional
        Figure size, by default (9, 8)
    output_path : str, optional
        Path to save the plot, by default None
    show_plot : bool, optional
        Whether to show the plot, by default True
    ax : matplotlib.axes.Axes, optional
        Axes to plot on, by default None
    line_width : int, optional
        Line width for the plot, by default 1
    title_size : int, optional
        Font size for the title, by default 16

    Raises
    ------
    ValueError
        If an invalid method is specified
    """
    method = method.lower()
    if method not in ["median", "mean"]:
        raise ValueError("Invalid method. Must be either 'median' or 'mean'")
    if period is None:
        period = 1 / candidate.freq

    # remove data points with flux = NaN
    time, flux, _, _ = remove_nondetections(
        photometry[0], photometry[1], photometry[2], photometry[3]
    )

    phase = (time / (period * 86400)) % 2  # period is converted from days to seconds

    # Create bins and calculate which bin each phase belongs to
    bin_edges = np.linspace(0, 1, num_bins + 1)
    bin_indices = np.digitize(phase, bin_edges) - 1  # -1 to make it zero-indexed

    method_flux_per_bin = np.zeros(num_bins)
    for i in range(num_bins):
        bin_fluxes = flux[bin_indices == i]

        if len(bin_fluxes) == 0 or np.all(np.isnan(bin_fluxes)):
            method_flux_per_bin[i] = 0
        else:
            if method == "median":
                method_flux_per_bin[i] = np.median(bin_fluxes)
            elif method == "mean":
                method_flux_per_bin[i] = np.mean(bin_fluxes)

    if ax is None:
        _, ax = plt.subplots(1, 1, figsize=figsize)

    ax.step(
        np.arange(2 * num_bins),
        np.concatenate((method_flux_per_bin, method_flux_per_bin)),
        where="mid",
        linewidth=line_width,
        color="black",
    )

    ax.set_xlabel("Bin Number")
    ax.set_ylabel("Median Flux")
    ax.set_title(
        f"{method.capitalize()} Flux vs Bin Number\nID: {int(candidate.psid)} Period: {24 / candidate.freq:.4f}",
        fontsize=title_size,
    )
    plt.grid(True)

    if output_path:
        output_dir = os.path.dirname(output_path)
        os.makedirs(output_dir, exist_ok=True)
        plt.savefig(output_path, dpi=300)

    if show_plot:
        plt.show()
        plt.close()


# just a wrapper with method='median' for plot_photometry_stat_per_bin
def plot_folded_photometry_median_per_bin(
    candidate: VariabilityCandidate,
    photometry: List[np.ndarray],
    period: float = None,
    num_bins: int = 20,
    figsize: tuple = (9, 8),
    output_path: str = None,
    show_plot: bool = True,
    ax=None,
    line_width: int = 1,
    title_size: int = 16,
):
    """
    Plot the median flux per bin for the folded lightcurve

    Parameters
    ----------
    candidate : VariabilityCandidate
        Variability candidate
    photometry : List[np.ndarray]
        Photometry data
    period : float, optional
        Period of the candidate, by default None
    num_bins : int, optional
        Number of bins to use for binning the data, by default 20
    figsize : tuple, optional
        Figure size, by default (9, 8)
    output_path : str, optional
        Path to save the plot, by default None
    show_plot : bool, optional
        Whether to show the plot, by default True
    ax : matplotlib.axes.Axes, optional
        Axes to plot on, by default None
    line_width : int, optional
        Line width for the plot, by default 1
    title_size : int, optional
        Font size for the title, by default 16

    Raises
    ------
    ValueError
        If an invalid method is specified
    """
    plot_folded_photometry_stat_per_bin(
        candidate,
        photometry,
        period,
        num_bins,
        "median",
        figsize,
        output_path,
        show_plot,
        ax,
        line_width,
        title_size,
    )


# just a wrapper with method='mean' for plot_photometry_stat_per_bin
def plot_folded_photometry_mean_per_bin(
    candidate: VariabilityCandidate,
    photometry: List[np.ndarray],
    period: float = None,
    num_bins: int = 20,
    figsize: tuple = (9, 8),
    output_path: str = None,
    show_plot: bool = True,
    ax=None,
    line_width: int = 1,
    title_size: int = 16,
):
    """
    Plot the mean flux per bin for the folded lightcurve

    Parameters
    ----------
    candidate : VariabilityCandidate
        Variability candidate
    photometry : List[np.ndarray]
        Photometry data
    period : float, optional
        Period of the candidate, by default None
    num_bins : int, optional
        Number of bins to use for binning the data, by default 20
    figsize : tuple, optional
        Figure size, by default (9, 8)
    output_path : str, optional
        Path to save the plot, by default None
    show_plot : bool, optional
        Whether to show the plot, by default True
    ax : matplotlib.axes.Axes, optional
        Axes to plot on, by default None
    line_width : int, optional
        Line width for the plot, by default 1
    title_size : int, optional
        Font size for the title, by default 16

    Raises
    ------
    ValueError
        If an invalid method is specified
    """
    plot_folded_photometry_stat_per_bin(
        candidate,
        photometry,
        period,
        num_bins,
        "mean",
        figsize,
        output_path,
        show_plot,
        ax,
        line_width,
        title_size,
    )



import os
from typing import Tuple, List, Union
import numpy as np
import h5py
from astropy.coordinates import SkyCoord

import fpw
import paramiko
import requests
from scp import SCPClient, SCPException
from penquins import Kowalski

from zvartools.spatial import get_field_id, great_circle_distance
from zvartools.enums import FILTER2IDX
from zvartools.lightcurves import cleanup_lightcurve, adu_to_jansky, freq_grid
from zvartools.candidate import (
    VariabilityCandidate,
    add_gaia_xmatch_to_candidates,
    add_ps1_xmatch_to_candidates,
    add_allwise_xmatch_to_candidates,
    add_2mass_xmatch_to_candidates,
    import_from_parquet,
)
from zvartools.enums import ALLOWED_BANDS, FILTER2IDX


class BaseDataSource:
    @staticmethod
    def add_kowalski(**kwargs):
        """
        Add a Kowalski instance if the required parameters are provided

        Parameters
        ----------
        kowalski_protocol: str
            The protocol to use to connect to Kowalski
        kowalski_host: str
            The host to connect to
        kowalski_port: int
            The port to connect to
        kowalski_token: str
            The token to use to authenticate, if provided, it will take precedence over kowalski_user and kowalski_password
        kowalski_user: str
            The user to use to authenticate, if not using a token
        kowalski_password: str
            The password to use to authenticate, if not using a token
        verbose: bool
            Whether to print debug information when interacting with Kowalski

        Returns
        -------
        Kowalski or None
            A Kowalski instance if the required parameters are provided, None otherwise
        """
        if (
            kwargs.get("kowalski_protocol")
            and kwargs.get("kowalski_host")
            and kwargs.get("kowalski_port")
        ):
            if kwargs.get("kowalski_token"):
                k = Kowalski(
                    protocol=kwargs.get("kowalski_protocol"),
                    host=kwargs.get("kowalski_host"),
                    port=443,
                    token=kwargs.get("kowalski_token"),
                    verbose=kwargs.get("verbose", False),
                    timeout=600,
                )
            elif kwargs.get("kowalski_user") and kwargs.get("kowalski_password"):
                k = Kowalski(
                    protocol=kwargs.get("kowalski_protocol"),
                    host=kwargs.get("kowalski_host"),
                    port=kwargs.get("kowalski_port"),
                    user=kwargs.get("kowalski_user"),
                    password=kwargs.get("kowalski_password"),
                    verbose=kwargs.get("verbose", False),
                    timeout=600,
                )
            else:
                raise ValueError("Kowalski credentials not provided")
            return k
        return None

    def __init__(
        self,
        local_lightcurve_path: str,
        local_period_path: str,
        **kwargs,
    ):
        self.local_lightcurve_path = local_lightcurve_path
        self.local_period_path = local_period_path
        self.kwargs = kwargs
        self.verbose = kwargs.get("verbose", False)
        self.secondary = kwargs.get("secondary", True)

        self.implements = {
            "check_file_availability",
            "get_files",
        }
        self.kowalski = self.add_kowalski(**kwargs)

    def check_file_availability(
        self, data: Tuple[int, int, int, str], type: str = "lightcurve"
    ) -> Tuple[List, List]:
        """
        Check if the files are available locally, if not, return the missing files

        Parameters
        ----------
        data: Tuple[int, int, int, str]
            The data to check, a list of tuples with field, ccd, quad, band
        type: str
            The type of data to check, either lightcurve or period

        Returns
        -------
        Tuple[List, List]
            A tuple with two lists, the first one containing the available files, the second one the missing files
        """
        base_dir, prefix = None, None
        if type == "lightcurve":
            base_dir, prefix = self.local_lightcurve_path, "data"
        elif type == "period":
            base_dir, prefix = self.local_period_path, "fpw"
        else:
            raise ValueError(f"Type {type} is not supported")

        available, missing = [], []

        for field, ccd, quad, band in data:
            if not self.secondary and field > 1000:
                continue
            filename = (
                f"{field:04d}/{prefix}_{field:04d}_{ccd:02d}_{quad:01d}_z{band}.h5"
            )
            if not os.path.isfile(f"{base_dir}/{filename}"):
                missing.append((field, ccd, quad, band))
            else:
                available.append((field, ccd, quad, band))

        return available, missing

    def get_files(self, data: Tuple[int, int, int, str], type="lightcurve") -> List:
        """
        Get the files locally, if they are not available, download them

        Parameters
        ----------
        data: Tuple[int, int, int, str]
            The data to check, a list of tuples with field, ccd, quad, band
        type: str
            The type of data to check, either lightcurve or period

        Returns
        -------
        List
            A list of the available files
        """
        available, missing = self.check_file_availability(data, type)
        if len(missing) > 0:
            if "download_files" in self.implements:
                if self.verbose:
                    print(f"Downloading {len(missing)} missing files")

                self.download_files(missing, type)
                available, missing = self.check_file_availability(data, type)

        if len(missing) > 0 and self.verbose:
            print(f"Missing {len(missing)} {type} files")

        return available

    def download_files(
        self, data: Tuple[int, int, int, str], type="lightcurve"
    ) -> List:
        """
        Download the files from the remote server (only implemented in the child classes)

        Parameters
        ----------
        data: Tuple[int, int, int, str]
            The data to download, a list of tuples with field, ccd, quad, band
        type: str
            The type of data to download, either lightcurve or period

        Raises
        ------
        NotImplementedError
            This method should be implemented in the child classes
        """
        raise NotImplementedError(
            "This method should be implemented in the child classes"
        )

    @staticmethod
    def read_lightcurve_by_idx(
        idx: int,
        sources_data: h5py.Dataset,
        times: np.array,
        band: str,
        zeropoints: np.array,
        exptimes: np.array,
    ) -> List:
        """
        Read the lightcurve data for a given source index

        Parameters
        ----------
        idx: int
            The index of the source
        sources_data: h5py.Dataset
            The dataset containing the sources data
        times: np.array
            The times of the exposures
        band: str
            The band of the exposures
        zeropoints: np.array
            The zeropoints of the exposures
        exptimes: np.array
            The exposure times

        Returns
        -------
        List
            A list of the photometry data, of shape (time, flux, flux_err, band) x nb_exposures
        """
        nb_exposures = len(times)
        start, end = idx * nb_exposures, (idx + 1) * nb_exposures

        raw_photometry = sources_data[start:end]

        photometry = []
        for i in range(nb_exposures):
            if (
                int(raw_photometry[i][2]) == 1
            ):  # if the source is not detected in this exposure
                photometry.append(
                    [
                        times[i],
                        np.nan,
                        float(raw_photometry[i][1]),
                        FILTER2IDX[band],
                    ]
                )
            else:  # if the source is detected in this exposure
                flux_jy, flux_err_jy = adu_to_jansky(
                    float(raw_photometry[i][0]),
                    float(raw_photometry[i][1]),
                    exptimes[i],
                    zeropoints[i],
                )
                # we convert to microJy
                photometry.append(
                    [
                        times[i],
                        flux_jy * 1e6,
                        flux_err_jy * 1e6,
                        FILTER2IDX[band],
                    ]
                )

        time, flux, flux_err = (
            np.array([x[0] for x in photometry]),
            np.array([x[1] for x in photometry]),
            np.array([x[2] for x in photometry]),
        )
        photometry = list(zip(time, flux, flux_err, [x[3] for x in photometry]))

        return photometry

    @staticmethod
    def read_period_by_idx(idx: int, dataset: h5py.File) -> Tuple:
        """
        Read the period data for a given source index

        Parameters
        ----------
        idx: int
            The index of the source
        dataset: h5py.File
            The dataset containing the period data

        Returns
        -------
        Tuple
            A tuple with the best frequency, significance, best M and valid flag
        """
        try:
            freqs = dataset["bestFreqs"][idx * 50 * 3 : (idx + 1) * 50 * 3]
            sigs = dataset["significance"][idx * 50 * 3 : (idx + 1) * 50 * 3]
            valid = dataset["valid"][idx]
            max_sig = np.max(sigs)
            max_sig_idx = np.argmax(sigs)
            best_M = 20 if max_sig_idx < 50 else 10 if max_sig_idx < 100 else 5
            best_freq = freqs[max_sig_idx]
            return best_freq, max_sig, best_M, valid
        except Exception as e:
            print(f"Error reading period for {idx}: {e}")
            return None, None, None, None

    @staticmethod
    def process_lightcurves(lightcurve_per_psid: dict) -> dict:
        """
        Process the lightcurves for each source, cleaning them up and sorting them by time

        Parameters
        ----------
        lightcurve_per_psid: dict
            A dictionary with a list of lightcurves for each source

        Returns
        -------
        dict
            A dictionary with the cleaned up lightcurves for each source
        """

        for id, photometry in lightcurve_per_psid.items():
            if not photometry:
                print(f"No photometry for {id} found in any file")
                continue
            lightcurve_per_psid[id] = np.array(sorted(photometry, key=lambda x: x[0])).T
            lightcurve_per_psid[id] = cleanup_lightcurve(*lightcurve_per_psid[id])
            lightcurve_per_psid[id] = np.array(lightcurve_per_psid[id], order="C")

        return lightcurve_per_psid

    def cone_search(
        self,
        targets: Union[list, tuple],
        radius: float = 2.0,
        max_matches: Union[int, None] = None,
    ):
        """
        Perform a cone search around the given targets

        Parameters
        ----------
        targets: list
            A list of targets, each target can be a tuple, list, dict, SkyCoord or object with ra and dec attributes
        radius: float
            The radius of the cone search in arcseconds
        max_matches: int
            The maximum number of matches to return

        Returns
        -------
        dict, dict
            A dictionary with the sources found for each target, a dictionary with the lightcurves for each source
        """
        fieldccdquad2radec = {}
        if isinstance(targets, (list, tuple)) and all(
            isinstance(t, (list, tuple)) and len(t) == 2 for t in targets
        ):
            positions = targets
        elif isinstance(targets, (list, tuple)) and all(
            isinstance(t, SkyCoord) for t in targets
        ):
            positions = [(t.ra.deg, t.dec.deg) for t in targets]
        elif isinstance(targets, (list, tuple)) and all(
            isinstance(t, dict) and "ra" in t and "dec" in t for t in targets
        ):
            positions = [(t["ra"], t["dec"]) for t in targets]
        # if its a python class with ra and dec attributes
        elif isinstance(targets, (list, tuple)) and all(
            isinstance(t, object)
            and isinstance(getattr(t, "ra", None), float)
            and isinstance(getattr(t, "dec", None), float)
            for t in targets
        ):
            positions = [(t.ra, t.dec) for t in targets]
        else:
            raise ValueError(
                "Targets should be a list of tuples, lists, SkyCoord objects, dicts or objects with ra and dec attributes"
            )

        for ra, dec in positions:
            field_ccd_quads = get_field_id(ra, dec, radius)

            for field, ccd, quad in field_ccd_quads:
                if (field, ccd, quad) not in fieldccdquad2radec:
                    fieldccdquad2radec[(field, ccd, quad)] = []
                fieldccdquad2radec[(field, ccd, quad)].append((ra, dec))

        data_to_recover = []
        for (field, ccd, quad), radecs in fieldccdquad2radec.items():
            for ra, dec in radecs:
                for band in ["g", "r"]:
                    data_to_recover.append((field, ccd, quad, band))

        available_lightcurves = self.get_files(data_to_recover, type="lightcurve")
        available_periods = self.get_files(data_to_recover, type="period")

        radec2sources = {(ra, dec): [] for ra, dec in positions}
        lightcurve_per_psid = {}
        periods_per_psid = {}
        for field, ccd, quad, band in available_lightcurves:
            matched_idx = set()
            lightcurve_file_path = f"{self.local_lightcurve_path}/{field:04d}/data_{field:04d}_{ccd:02d}_{quad:01d}_z{band}.h5"
            with h5py.File(lightcurve_file_path, "r") as f:
                data = f["data"]
                sources: h5py.Dataset = data["sources"]

                for ra, dec in fieldccdquad2radec[(field, ccd, quad)]:
                    distances = great_circle_distance(
                        ra, dec, sources["ra"], sources["decl"]
                    )
                    # TODO: if max_matches is defined, limit the nb of results returned.
                    matches = distances < radius / 3600
                    if max_matches:
                        # sort by distance and take the first max_matches
                        matches = np.argsort(distances)[0:max_matches]
                    for i in np.where(matches)[0]:
                        psid = sources["gaia_id"][i]
                        radec2sources[(ra, dec)].append(
                            {
                                "psid": psid,
                                "ra": sources["ra"][i],
                                "dec": sources["decl"][i],
                                "field": field,
                                "ccd": ccd,
                                "quad": quad,
                                "band": band,
                            }
                        )
                        if psid not in lightcurve_per_psid:
                            lightcurve_per_psid[psid] = []
                        matched_idx.add((i, psid))

                for i, psid in matched_idx:
                    lc = self.read_lightcurve_by_idx(
                        i,
                        data["sourcedata"],
                        data["exposures"]["bjd"],
                        band,
                        data["exposures"]["mzpsci"],
                        data["exposures"]["exptime"],
                    )
                    lightcurve_per_psid[psid].extend(lc)

            if (field, ccd, quad, band) not in available_periods:
                print(
                    f"Period file for {field:04d}_{ccd:02d}_{quad:01d}_z{band} not found"
                )
                continue
            periods_file_path = f"{self.local_period_path}/{field:04d}/fpw_{field:04d}_{ccd:02d}_{quad:01d}_z{band}.h5"
            with h5py.File(periods_file_path, "r") as dataset:
                psids = np.array(dataset["psids"])
                assert len(dataset["bestFreqs"]) == len(dataset["psids"]) * 50 * 3

                for _, psid in matched_idx:
                    idx = np.where(psids == psid)[0]
                    if len(idx) == 0:
                        continue

                    freq, sig, best_M, valid = self.read_period_by_idx(idx[0], dataset)
                    if (psid, band) not in periods_per_psid:
                        periods_per_psid[(psid, band)] = []
                    periods_per_psid[(psid, band)].append(
                        {
                            "freq": freq,
                            "significance": sig,
                            "best_M": best_M,
                            "valid": valid,
                        }
                    )

        radec2sources = {
            (ra, dec): list(match) for (ra, dec), match in radec2sources.items()
        }
        for key in radec2sources:
            radec2sources[key] = sorted(radec2sources[key], key=lambda x: x["psid"])
            # for each, add the period data if available
            for i in range(len(radec2sources[key])):
                psid = radec2sources[key][i]["psid"]
                if (psid, radec2sources[key][i]["band"]) in periods_per_psid:
                    radec2sources[key][i] = {
                        **radec2sources[key][i],
                        **periods_per_psid[(psid, radec2sources[key][i]["band"])][0],
                    }
                # convert them into VariabilityCandidate objects
                radec2sources[key][i] = VariabilityCandidate(
                    psid=psid,
                    ra=radec2sources[key][i]["ra"],
                    dec=radec2sources[key][i]["dec"],
                    band=radec2sources[key][i].get("band", None),
                    freq=radec2sources[key][i].get("freq", None),
                    fap=radec2sources[key][i].get("significance", None),
                    best_M=radec2sources[key][i].get("best_M", None),
                    field=radec2sources[key][i].get("field", None),
                    ccd=radec2sources[key][i].get("ccd", None),
                    quad=radec2sources[key][i].get("quad", None),
                    valid=radec2sources[key][i].get("valid", None),
                )

        lightcurve_per_psid = self.process_lightcurves(lightcurve_per_psid)

        return radec2sources, lightcurve_per_psid

    @staticmethod
    def psid2radec(psid):
        psid = str(int(psid))
        ZH = 0.0083333
        ZID = str(f"{int(str(psid[0:5]).lstrip('0'))}.{psid[14:18]}")
        dec = (float(ZID) * ZH) - 90.0
        ra_str = str(f"{int(str(psid[5:8]).lstrip('0'))}.{psid[8:14]}")
        ra = float(ra_str)
        return ra, dec

    def psid_search(self, pstargets: Union[list, tuple]):
        """
        Search for the given Pan-STARRs targets

        Parameters
        ----------
        pstargets: list
            A list of Pan-STARRs ids

        Returns
        -------
        dict, dict
            A dictionary with the sources found for each target, a dictionary with the lightcurves for each source
        """
        if not isinstance(pstargets, (list, tuple)):
            raise ValueError(
                "Pan-STARRs targets should be a list of tuples, lists or dicts"
            )
        if not all(isinstance(t, (int, str, float)) for t in pstargets):
            raise ValueError(
                "Pan-STARRs targets should be a list of integers or strings"
            )

        lightcurve_per_psid = {}
        metadata_per_psidband = {}
        # create a mapper from field, ccd, quad to psid
        fieldccdquad2psid = {}
        data_to_recover = []
        for psid in pstargets:
            ra, dec = self.psid2radec(psid)
            field_ccd_quads = get_field_id(ra, dec, 2.0)
            for field, ccd, quad in field_ccd_quads:
                if (field, ccd, quad) not in fieldccdquad2psid:
                    fieldccdquad2psid[(field, ccd, quad)] = set()
                    for band in ["g", "r"]:
                        data_to_recover.append((field, ccd, quad, band))
                fieldccdquad2psid[(field, ccd, quad)].add(psid)

        available_lightcurves = self.get_files(data_to_recover, type="lightcurve")
        available_periods = self.get_files(data_to_recover, type="period")

        for field, ccd, quad, band in available_lightcurves:
            lightcurve_file_path = f"{self.local_lightcurve_path}/{field:04d}/data_{field:04d}_{ccd:02d}_{quad:01d}_z{band}.h5"
            with h5py.File(lightcurve_file_path, "r") as f:
                data = f["data"]
                sources: h5py.Dataset = data["sources"]
                for psid in fieldccdquad2psid[(field, ccd, quad)]:
                    idx = np.where(sources["gaia_id"][:] == psid)[0]
                    if len(idx) == 0:
                        continue
                    idx = idx[0]
                    lc = self.read_lightcurve_by_idx(
                        idx,
                        data["sourcedata"],
                        data["exposures"]["bjd"],
                        band,
                        data["exposures"]["mzpsci"],
                        data["exposures"]["exptime"],
                    )
                    if psid not in lightcurve_per_psid:
                        lightcurve_per_psid[psid] = []
                    if (psid, band) not in metadata_per_psidband:
                        metadata_per_psidband[(psid, band)] = {
                            "psid": psid,
                            "ra": sources["ra"][idx],
                            "dec": sources["decl"][idx],
                            "field": field,
                            "ccd": ccd,
                            "quad": quad,
                            "band": band,
                        }
                    lightcurve_per_psid[psid].extend(lc)

            if (field, ccd, quad, band) not in available_periods:
                print(
                    f"Period file for {field:04d}_{ccd:02d}_{quad:01d}_z{band} not found"
                )
                continue
            periods_file_path = f"{self.local_period_path}/{field:04d}/fpw_{field:04d}_{ccd:02d}_{quad:01d}_z{band}.h5"
            with h5py.File(periods_file_path, "r") as dataset:
                psids = np.array(dataset["psids"])
                assert len(dataset["bestFreqs"]) == len(dataset["psids"]) * 50 * 3

                for psid in fieldccdquad2psid[(field, ccd, quad)]:
                    idx = np.where(psids == psid)[0]
                    if len(idx) == 0:
                        continue

                    freq, sig, best_M, valid = self.read_period_by_idx(idx[0], dataset)
                    metadata_per_psidband[(psid, band)] = {
                        **metadata_per_psidband[(psid, band)],
                        "freq": freq,
                        "significance": sig,
                        "best_M": best_M,
                        "valid": valid,
                    }

        lightcurve_per_psid = self.process_lightcurves(lightcurve_per_psid)

        # then create the VariabilityCandidate objects
        candidates = {}
        for (psid, band), metadata in metadata_per_psidband.items():
            if psid not in candidates:
                candidates[psid] = []
            candidates[psid].append(
                VariabilityCandidate(
                    psid=psid,
                    ra=metadata["ra"],
                    dec=metadata["dec"],
                    band=metadata["band"],
                    freq=metadata["freq"],
                    fap=metadata["significance"],
                    best_M=metadata["best_M"],
                    field=metadata["field"],
                    ccd=metadata["ccd"],
                    quad=metadata["quad"],
                    valid=metadata["valid"],
                )
            )

        return candidates, lightcurve_per_psid

    def get_candidates_lightcurves(self, candidates: List[VariabilityCandidate]):
        """
        Get the lightcurves for the given candidates

        Parameters
        ----------
        candidates: list
            A list of VariabilityCandidate objects

        Returns
        -------
        dict
            A dictionary with the cleaned up lightcurves for each source
        """
        if not isinstance(candidates, (list, tuple)) and all(
            isinstance(c, VariabilityCandidate) for c in candidates
        ):
            raise ValueError(
                "Candidates should be a list of VariabilityCandidate objects"
            )

        data_to_recover = []
        for candidate in candidates:
            if candidate.band is None:
                for band in ["g", "r"]:
                    data_to_recover.append(
                        (candidate.field, candidate.ccd, candidate.quad, band)
                    )
            else:
                data_to_recover.append(
                    (candidate.field, candidate.ccd, candidate.quad, candidate.band)
                )

        available_lightcurves = self.get_files(data_to_recover, type="lightcurve")

        lightcurve_per_psid = {}

        for field, ccd, quad, band in available_lightcurves:
            lightcurve_file_path = f"{self.local_lightcurve_path}/{field:04d}/data_{field:04d}_{ccd:02d}_{quad:01d}_z{band}.h5"
            with h5py.File(lightcurve_file_path, "r") as f:
                data = f["data"]
                sources: h5py.Dataset = data["sources"]
                for candidate in candidates:
                    idx = np.where(sources["gaia_id"][:] == candidate.psid)[0]
                    if len(idx) == 0:
                        continue
                    idx = idx[0]
                    lc = self.read_lightcurve_by_idx(
                        idx,
                        data["sourcedata"],
                        data["exposures"]["bjd"],
                        band,
                        data["exposures"]["mzpsci"],
                        data["exposures"]["exptime"],
                    )
                    if candidate.psid not in lightcurve_per_psid:
                        lightcurve_per_psid[candidate.psid] = []
                    lightcurve_per_psid[candidate.psid].extend(lc)

        lightcurve_per_psid = self.process_lightcurves(lightcurve_per_psid)

        return lightcurve_per_psid

    def validate_add_data_parameters(
        self, candidates: List[VariabilityCandidate], catalog_name: str
    ):
        """
        Validate the parameters for the add_data_to_candidates methods

        Parameters
        ----------
        candidates: list
            A list of VariabilityCandidate objects
        catalog_name: str
            The name of the catalog to add data from

        Raises
        ------
        ValueError
            If the candidates are not a list of VariabilityCandidate objects
        ValueError
            If the Kowalski instance is not provided
        """
        if not isinstance(candidates, (list, tuple)) and all(
            isinstance(c, VariabilityCandidate) for c in candidates
        ):
            raise ValueError(
                "Candidates should be a list of VariabilityCandidate objects"
            )

        if self.kowalski is None:
            raise ValueError(
                f"Kowalski instance not provided, required to add {catalog_name} data to candidates"
            )

    def add_ps1_data_to_candidates(self, candidates: List[VariabilityCandidate]):
        """
        Add Pan-STARRs data to the given candidates

        Parameters
        ----------
        candidates: list
            A list of VariabilityCandidate objects

        Returns
        -------
        list
            A list of the same candidates, with the Pan-STARRs data added
        """
        self.validate_add_data_parameters(candidates, "Pan-STARRs")
        return add_ps1_xmatch_to_candidates(self.kowalski, candidates)

    def add_gaia_data_to_candidates(
        self, candidates: List[VariabilityCandidate], radius: float = 3.0
    ):
        """
        Add Gaia data to the given candidates

        Parameters
        ----------
        candidates: list
            A list of VariabilityCandidate objects
        radius: float
            The radius of the cone search in arcseconds

        Returns
        -------
        list
            A list of the same candidates, with the Gaia data added
        """
        self.validate_add_data_parameters(candidates, "Gaia")
        return add_gaia_xmatch_to_candidates(self.kowalski, candidates, radius)

    def add_allwise_data_to_candidates(
        self, candidates: List[VariabilityCandidate], radius: float = 3.0
    ):
        """
        Add AllWISE data to the given candidates

        Parameters
        ----------
        candidates: list
            A list of VariabilityCandidate objects
        radius: float
            The radius of the cone search in arcseconds

        Returns
        -------
        list
            A list of the same candidates, with the AllWISE data added
        """
        self.validate_add_data_parameters(candidates, "AllWISE")
        return add_allwise_xmatch_to_candidates(self.kowalski, candidates, radius)

    def add_2mass_data_to_candidates(
        self, candidates: List[VariabilityCandidate], radius: float = 3.0
    ):
        """
        Add 2MASS data to the given candidates

        Parameters
        ----------
        candidates: list
            A list of VariabilityCandidate objects
        radius: float
            The radius of the cone search in arcseconds

        Returns
        -------
        list
            A list of the same candidates, with the 2MASS data added
        """
        self.validate_add_data_parameters(candidates, "2MASS")
        return add_2mass_xmatch_to_candidates(self.kowalski, candidates, radius)

    def load_variability_candidate_dataset(
        self, fields: List[int], bands: List[str] = ALLOWED_BANDS, path: str = None
    ) -> List[VariabilityCandidate]:
        """
        Load the variability candidates dataset

        Parameters
        ----------
        fields: list
            A list of fields to load
        bands: list
            A list of bands to load
        max_candidates: int
            The maximum number of candidates to load
        path: str
            The path to the datasets

        Returns
        -------
        dict
            A dictionary with the loaded candidates
        """
        if path is None:
            raise ValueError("Path to the datasets is required")

        candidates = []
        # check if a directory exists at that path
        if not os.path.isdir(path):
            raise ValueError(f"Path {path} does not exist")

        for field in fields:
            for band in bands:
                filename = f"{path}/{field:04d}/fpw_{field:04d}_z{band}.parquet"
                if not os.path.isfile(filename):
                    if self.verbose:
                        print(f"File {filename} not found")
                    continue

                candidates_temp = import_from_parquet(filename)
                candidates.extend(candidates_temp)

        return candidates

    def load_all_objects_from_matchfile(self, field, ccd, quad, band):
        """
        Load all objects from a given path to an h5 file
        """
        filename = f"{self.local_lightcurve_path}/{field:04d}/data_{field:04d}_{ccd:02d}_{quad:01d}_z{band}.h5"
        if not os.path.isfile(filename):
            raise ValueError(f"File {filename} not found")

        candidates = []
        with h5py.File(filename, "r") as f:
            data = f["data"]
            sources: h5py.Dataset = data["sources"]
            for i in range(len(sources)):
                psid = sources["gaia_id"][i]
                ra = sources["ra"][i]
                dec = sources["decl"][i]
                candidates.append(
                    VariabilityCandidate(
                        psid=psid,
                        ra=ra,
                        dec=dec,
                        valid=True,
                        freq=None,
                        fap=None,
                        best_M=None,
                        field=field,
                        ccd=ccd,
                        quad=quad,
                        band=band,
                    )
                )

        return candidates

    def fpw_periodogram(
        self,
        lightcurve,
        band="r",
        nb_periods=1,
        period_unit="hours",
        binsize=5,
        freq_min=None,
        freq_max=None,
    ):
        """
        Compute the periodogram using the FPW algorithm

        Parameters
        ----------
        lightcurve: np.ndarray
            The lightcurve data, of shape (time, flux, flux_err, band)
        band: str
            The band to compute the periodogram for
        nb_periods: int
            The number of periods to return
        period_unit: str
            The unit to return the period in: minutes, hours, days
        binsize: int
            The binsize to use
        freq_min: float
            The minimum frequency to consider
        freq_max: float
            The maximum frequency to consider

        Returns
        -------
        Tuple
            A tuple with the N best periods, their significances and the full periodogram
        """
        if band not in ALLOWED_BANDS:
            raise ValueError(
                f"Band {band} is not supported, must be one of {ALLOWED_BANDS}"
            )
        if nb_periods < 1:
            raise ValueError("Number of periods must be greater than 0")
        if period_unit not in ["minutes", "hours", "days"]:
            raise ValueError("Period unit must be one of minutes, hours, days")

        if period_unit == "minutes":
            multiplier = 60
        elif period_unit == "hours":
            multiplier = 3600
        elif period_unit == "days":
            multiplier = 86400

        lc = lightcurve.copy()
        lc = lc[:, lc[3] == FILTER2IDX[band]]

        # we drop the non detections
        lc = lc[~np.isnan(lc[:, 1])]
        # we make sure all arrays are C-contiguous, required by FPW
        time, flux, flux_err = (np.ascontiguousarray(x) for x in lc[:3])

        # we center the flux on 0 by subtracting the median
        flux -= np.nanmedian(flux)

        f_grid = freq_grid(time, fmin=freq_min, fmax=freq_max)

        fpw_pgram = fpw.run_fpw(time, flux, flux_err, f_grid, binsize)
        fpw_pgram[np.isnan(fpw_pgram)] = 0

        nb_periods = min(len(fpw_pgram[fpw_pgram > 0]), nb_periods)

        # return the periods and their significance, ordered by significance desc, up to nb_periods
        periods = []
        significances = []
        for i in np.argsort(fpw_pgram)[::-1][:nb_periods]:
            periods.append(1 / f_grid[i] / multiplier)
            significances.append(fpw_pgram[i])

        return periods, significances, fpw_pgram


class RemoteDataSource(BaseDataSource):
    def __init__(
        self,
        local_lightcurve_path,
        local_period_path,
        remote_lightcurve_path,
        remote_period_path,
        ssh_host,
        ssh_user,
        ssh_password=None,
        ssh_key_filename=None,
        **kwargs,
    ):
        super().__init__(local_lightcurve_path, local_period_path, **kwargs)
        self.remote_lightcurve_path = remote_lightcurve_path
        self.remote_period_path = remote_period_path

        self.ssh_client = paramiko.SSHClient()
        self.ssh_client.load_system_host_keys()
        if ssh_key_filename:
            self.ssh_client.connect(
                ssh_host, username=ssh_user, key_filename=ssh_key_filename
            )
        else:
            self.ssh_client.connect(ssh_host, username=ssh_user, password=ssh_password)
        self.implements.add("download_files")

    def download_files(
        self, data: Tuple[int, int, int, str], type: str = "lightcurve"
    ) -> List:
        """
        Download the files from the remote server

        Parameters
        ----------
        data: Tuple[int, int, int, str]
            The data to download, a list of tuples with field, ccd, quad, band
        type: str
            The type of data to download, either lightcurve or period

        Returns
        -------
        List
            A list of the downloaded files
        """
        scp_client = SCPClient(self.ssh_client.get_transport())

        local_base_dir, remote_base_dir, prefix = None, None, None
        if type == "lightcurve":
            local_base_dir, remote_base_dir, prefix = (
                self.local_lightcurve_path,
                self.remote_lightcurve_path,
                "data",
            )
        elif type == "period":
            local_base_dir, remote_base_dir, prefix = (
                self.local_period_path,
                self.remote_period_path,
                "fpw",
            )
        else:
            raise ValueError(f"Type {type} is not supported")

        downloaded = []

        for field, ccd, quad, band in data:
            filename = (
                f"{field:04d}/{prefix}_{field:04d}_{ccd:02d}_{quad:01d}_z{band}.h5"
            )
            remote_filename = f"{remote_base_dir}/{filename}"
            outdir = f"{local_base_dir}/{field:04d}"
            if not os.path.isdir(outdir):
                os.makedirs(outdir)

            # if we already have the file, we skip it
            if os.path.isfile(f"{local_base_dir}/{filename}"):
                continue

            # check if the file exists on the remote server
            _, stdout, stderr = self.ssh_client.exec_command(f"ls {remote_filename}")
            if not stdout.read():
                continue
            if stderr.read():
                if self.verbose:
                    print(
                        f"Could not check if {remote_filename} exists: {stderr.read()}"
                    )
                continue

            if self.verbose:
                print(f"Downloading {remote_filename} to {outdir}")
            try:
                scp_client.get(remote_filename, outdir)
            except SCPException as e:
                if self.verbose:
                    print(f"Could not download {remote_filename}: {e}")

            if self.verbose:
                print(
                    f"File size: {os.path.getsize(f'{local_base_dir}/{filename}') / 1024 / 1024:.2f} MB"
                )

            downloaded.append((field, ccd, quad, band))

        scp_client.close()

        return downloaded


class LocalDataSource(BaseDataSource):
    def __init__(self, local_lightcurve_path, local_period_path, **kwargs):
        super().__init__(local_lightcurve_path, local_period_path, **kwargs)

    def download_files(
        self, data: Tuple[int, int, int, str], type="lightcurve"
    ) -> List:
        """
        Download the files from the local server, not implemented in this class

        Raises
        ------
        NotImplementedError
            LocalDataSource does not support downloading files
        """
        raise NotImplementedError("LocalDataSource does not support downloading files")


class APIDataSource(BaseDataSource):
    def __init__(
        self,
        local_lightcurve_path,
        local_period_path,
        api_url,
        api_user,
        api_password,
        **kwargs,
    ):
        super().__init__(local_lightcurve_path, local_period_path, **kwargs)
        self.api_url = str(api_url).rstrip("/")
        self.api_user = api_user
        self.api_password = api_password
        self.implements.add("download_files")

    def test_api_connection(self):
        """
        Test the connection to the API

        Returns
        -------
        bool
            True if the connection is successful, False otherwise
        """
        try:
            response = requests.get(
                f"{self.api_url}/api/hello", auth=(self.api_user, self.api_password)
            )
            response.raise_for_status()
            return True
        except Exception as e:
            if self.verbose:
                print(f"Could not connect to the API: {e}")
            return False

    def download_files(
        self, data: Tuple[int, int, int, str], type: str = "lightcurve"
    ) -> List:
        """
        Download the files from an API

        Parameters
        ----------
        data: Tuple[int, int, int, str]
            The data to download, a list of tuples with field, ccd, quad, band
        type: str
            The type of data to download, either lightcurve or period

        Returns
        -------
        List
            A list of the downloaded files
        """
        local_base_dir, prefix = None, None
        if type == "lightcurve":
            local_base_dir, prefix, api_endpoint = (
                self.local_lightcurve_path,
                "data",
                "lightcurves",
            )
        elif type == "period":
            local_base_dir, prefix, api_endpoint = (
                self.local_period_path,
                "fpw",
                "periods",
            )
        else:
            raise ValueError(f"Type {type} is not supported")

        downloaded = []

        for field, ccd, quad, band in data:
            filename = (
                f"{field:04d}/{prefix}_{field:04d}_{ccd:02d}_{quad:01d}_z{band}.h5"
            )
            outdir = f"{local_base_dir}/{field:04d}"
            if not os.path.isdir(outdir):
                os.makedirs(outdir)

            # if we already have the file, we skip it
            if os.path.isfile(f"{local_base_dir}/{filename}"):
                continue

            if self.verbose:
                print(f"Downloading {filename} to {outdir}")

            response = requests.get(
                f"{self.api_url}/{api_endpoint}",
                params={"field": field, "ccd": ccd, "quad": quad, "filter": band},
                stream=True,
                auth=(self.api_user, self.api_password),
            )
            if response.status_code != 200:
                if self.verbose:
                    if response.status_code == 404:
                        print(f"File {filename} not found")
                    elif response.status_code == 401:
                        print(f"Unauthorized access to {filename}")
                    else:
                        print(f"Could not download {filename}: {response.text}")
                continue

            with open(f"{local_base_dir}/{filename}", "wb") as f:
                for chunk in response.iter_content(chunk_size=1024):
                    f.write(chunk)

            if self.verbose:
                print(
                    f"File size: {os.path.getsize(f'{local_base_dir}/{filename}') / 1024 / 1024:.2f} MB"
                )

            downloaded.append((field, ccd, quad, band))

        return downloaded



import glob
import os
from typing import Tuple
from multiprocessing import Pool, cpu_count
from tqdm import tqdm

import h5py
import numpy as np


def load_file(
    file: str,
) -> Tuple[
    str, Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]
]:
    """
    Load the periodicity (FPW) data from a file

    Parameters
    ----------
    file : str
        Path to the file

    Returns
    -------
    Tuple[str, Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]]
        File name and the data from the file
    """
    try:
        psids = np.array([], dtype=np.uint64)
        ratio_valid = np.array([])
        best_freqs = np.array([])
        significances = np.array([])
        ra = np.array([])
        dec = np.array([])

        with h5py.File(file, "r") as dataset:
            psids = np.array(dataset["psids"])
            ratio_valid = np.array(dataset["valid"])
            best_freqs = np.array(dataset["bestFreqs"])
            significances = np.array(dataset["significance"])
            # clean the significances
            np.nan_to_num(significances, nan=0, posinf=0, neginf=0, copy=False)
            ra = np.array(dataset["ra"])
            dec = np.array(dataset["dec"])

        # make it a tuple so we can return it
        return file, (psids, ra, dec, ratio_valid, best_freqs, significances)
    except FileNotFoundError:
        print(f"File {file} not found")
        return file, None
    except OSError:
        print(f"Error reading file {file}")
        return file, None
    except KeyError as e:
        print(f"Key not found in file {file}: {e}")
        return file, None
    except Exception as e:
        print(f"Error reading file {file}: {e}")
        return file, None


def get_ccd_quad_from_filename(file_name: str) -> Tuple[int, int]:
    """
    Get the CCD and quadrant numbers from the file name

    Parameters
    ----------
    file_name : str
        Name of the file

    Returns
    -------
    Tuple[int, int]
        CCD and quadrant numbers
    """
    ccd = int(file_name.split("_")[-3])
    quad = int(file_name.split("_")[-2])
    return ccd, quad


def load_field_periodicity_data_parallel(
    field: int, band: int, path: str
) -> Tuple[
    np.ndarray,
    np.ndarray,
    np.ndarray,
    np.ndarray,
    np.ndarray,
    np.ndarray,
    np.ndarray,
    np.ndarray,
]:
    """
    Load the periodicity data for a given field and band, loading the data in parallel

    Parameters
    ----------
    field : int
        Field number
    band : int
        Band number
    path : str
        Path to the data files

    Returns
    -------
    Tuple[
        np.ndarray,
        np.ndarray,
        np.ndarray,
        np.ndarray,
        np.ndarray,
        np.ndarray,
        np.ndarray,
        np.ndarray,
    ]
        PSIDs, RA, Dec, ratio_valid, best_freqs, significances, CCDs, quadrants
    """
    if not isinstance(field, (int, np.integer, str)):
        raise ValueError("Field must be an integer or string")
    if not isinstance(band, (int, np.integer, str)):
        raise ValueError("Band must be an integer or string")
    if path is None:
        raise ValueError("Path must be specified")

    path = os.path.join(
        os.path.abspath(path), f"{str(field).zfill(4)}", f"fpw_*_z{band}.h5"
    )

    files = list(set(glob.glob(path)))

    if not files:
        raise ValueError("No files found")

    # create an array of size len(files) to store the results
    data_per_file = np.empty(len(files), dtype=object)

    # create a pool of workers
    with Pool(min(8, cpu_count() - 2)) as pool:
        with tqdm(total=len(files), desc="Loading files") as pbar:
            for file_name, data in pool.imap_unordered(load_file, files):
                if data is None:
                    continue
                idx = files.index(file_name)
                data_per_file[idx] = data
                pbar.update(1)

    # remove the empty entries
    data_per_file = data_per_file[data_per_file != np.array(None)]

    if len(data_per_file) == 0:
        raise ValueError("No data found in files")

    # now we are left with a numpy array of arrays, i.e. a 3D array
    # reshape to a 2D array

    # create an empty array to store the results
    psids = np.empty(sum(len(data[0]) for data in data_per_file), dtype=np.uint64)
    ra = np.empty(sum(len(data[1]) for data in data_per_file))
    dec = np.empty(sum(len(data[2]) for data in data_per_file))
    ratio_valid = np.empty(sum(len(data[3]) for data in data_per_file))
    best_freqs = np.empty(sum(len(data[4]) for data in data_per_file))
    significances = np.empty(sum(len(data[5]) for data in data_per_file))

    # we also want arrays of ccd and quadrant numbers, to keep track of where the data comes from
    ccds = np.empty(sum(len(data[0]) for data in data_per_file), dtype=np.uint8)
    quads = np.empty(sum(len(data[0]) for data in data_per_file), dtype=np.uint8)

    current_idx_psids = 0
    current_idx_ra = 0
    current_idx_dec = 0
    current_idx_ratio_valid = 0
    current_idx_best_freqs = 0
    current_idx_significances = 0
    for i, data in tqdm(
        enumerate(data_per_file), desc="Processing data", total=len(data_per_file)
    ):
        # now we don't append to the arrays, but rather insert the data at the correct positions
        psids[current_idx_psids : current_idx_psids + len(data[0])] = data[0]
        ra[current_idx_ra : current_idx_ra + len(data[1])] = data[1]
        dec[current_idx_dec : current_idx_dec + len(data[2])] = data[2]
        ratio_valid[
            current_idx_ratio_valid : current_idx_ratio_valid + len(data[3])
        ] = data[3]
        best_freqs[current_idx_best_freqs : current_idx_best_freqs + len(data[4])] = (
            data[4]
        )
        significances[
            current_idx_significances : current_idx_significances + len(data[5])
        ] = data[5]

        file_name = files[i]
        ccd, quad = get_ccd_quad_from_filename(file_name)
        ccds[current_idx_psids : current_idx_psids + len(data[0])] = ccd
        quads[current_idx_psids : current_idx_psids + len(data[0])] = quad

        current_idx_psids += len(data[0])
        current_idx_ra += len(data[1])
        current_idx_dec += len(data[2])
        current_idx_ratio_valid += len(data[3])
        current_idx_best_freqs += len(data[4])
        current_idx_significances += len(data[5])

    if len(psids) == 0:
        raise ValueError("No data found in files")

    freqs = best_freqs.reshape((len(psids), 3, 50))
    sigs = significances.reshape((len(psids), 3, 50))

    return psids, ra, dec, ratio_valid, freqs, sigs, ccds, quads



import os
from typing import Union

import numpy as np
from numba import njit

CCD_LAYOUT_X = [
    -3.646513,
    -3.647394,
    -1.920848,
    -1.920383,
    -1.790386,
    -1.790817,
    -0.064115,
    -0.064099,
    0.062113,
    0.062129,
    1.788830,
    1.788400,
    1.918441,
    1.918905,
    3.645452,
    3.644571,
    -3.646416,
    -3.646708,
    -1.919998,
    -1.919844,
    -1.789454,
    -1.789597,
    -0.062733,
    -0.062727,
    0.061814,
    0.061819,
    1.788683,
    1.788540,
    1.918871,
    1.919025,
    3.645736,
    3.645443,
    -3.646562,
    -3.646270,
    -1.919698,
    -1.919852,
    -1.789413,
    -1.789270,
    -0.062544,
    -0.062549,
    0.062876,
    0.062871,
    1.789598,
    1.789741,
    1.919874,
    1.919720,
    3.646292,
    3.646584,
    -3.645853,
    -3.644972,
    -1.918842,
    -1.919306,
    -1.789367,
    -1.788937,
    -0.062651,
    -0.062666,
    0.063143,
    0.063128,
    1.789415,
    1.789845,
    1.919878,
    1.919413,
    3.645543,
    3.646424,
]

CCD_LAYOUT_Y = [
    -3.727898,
    -2.001758,
    -2.004785,
    -3.731333,
    -3.729368,
    -2.002803,
    -2.003812,
    -3.730512,
    -3.730976,
    -2.004276,
    -2.003269,
    -3.729834,
    -3.731505,
    -2.004957,
    -2.001932,
    -3.728073,
    -1.816060,
    -0.089749,
    -0.090622,
    -1.817335,
    -1.816611,
    -0.089881,
    -0.090172,
    -1.817035,
    -1.817472,
    -0.090609,
    -0.090319,
    -1.817048,
    -1.817584,
    -0.090871,
    -0.089998,
    -1.816309,
    0.090679,
    1.816989,
    1.818266,
    0.091552,
    0.091155,
    1.817884,
    1.818309,
    0.091446,
    0.090876,
    1.817739,
    1.817315,
    0.090586,
    0.091290,
    1.818003,
    1.816728,
    0.090417,
    2.002667,
    3.728808,
    3.732241,
    2.005694,
    2.003694,
    3.730258,
    3.731401,
    2.004701,
    2.003834,
    3.730533,
    3.729391,
    2.002826,
    2.004674,
    3.731221,
    3.727789,
    2.001648,
]

# maximal angular distance between the center of the field and its edge
# in radians
ADIST_MAX = 5.66 * np.pi / 180.0

# buffer for the CCD layout, to avoid edge effects
CCD_BUFFER = 0.03

DEGRA = np.pi / 180.0


class ZTFFieldData:
    fieldno = None
    ra_all = None
    dec_all = None

    def __init__(self):
        self.fieldno, self.ra_all, self.dec_all = self._get_fields_data()

    def _get_fields_data(self):
        """
        Load the ZTF field data from the file ZTF_Fields.txt

        Returns
        -------
        tuple
            fieldno, ra_all, dec_all
        """
        if self.fieldno is not None:
            return self.fieldno, self.ra_all, self.dec_all

        field_path = os.path.join(os.path.dirname(__file__), "./data/ZTF_Fields.txt")
        _fieldno, _ra_all, _dec_all = np.loadtxt(
            field_path, unpack=True, usecols=(0, 1, 2), dtype="int,float,float"
        )
        return _fieldno, _ra_all * DEGRA, _dec_all * DEGRA


ZTFFieldData = ZTFFieldData()


@njit
def fit_line(x: float, x0: float, y0: float, x1: float, y1: float) -> float:
    """
    Fit a linear function connecting two points (x0,y0) and (x1,y1)
    and evaluate its value at point x.

    Parameters
    ----------
    x : float
        x-coordinate of the point
    x0 : float
        x-coordinate of the first point
    y0 : float
        y-coordinate of the first point
    x1 : float
        x-coordinate of the second point
    y1 : float
        y-coordinate of the second point

    Returns
    -------
    float
        y-coordinate of the point on the line
    """

    return (y1 - y0) * (x - x0) / (x1 - x0) + y0


@njit
def ortographic_projection(ra: float, dec: float, ra0: float, dec0: float) -> tuple:
    """
    Calculate the ortographic projection onto the (x,y) tangent plane
    with the origin at (ra0,dec0)
    See: https://en.wikipedia.org/wiki/Orthographic_map_projection

    Parameters
    ----------
    ra : float
        Right ascension of the object
    dec : float
        Declination of the object
    ra0 : float
        Right ascension of the origin
    dec0 : float
        Declination of the origin

    Returns
    -------
    tuple
        x, y coordinates of the object
    """

    x = -np.cos(dec) * np.sin(ra - ra0)
    y = np.cos(dec0) * np.sin(dec) - np.sin(dec0) * np.cos(dec) * np.cos(ra - ra0)

    x *= 180.0 / np.pi
    y *= 180.0 / np.pi

    return x, y


@njit
def inside_polygon(xp: float, yp: float, x: list, y: list) -> tuple:
    """
    Check if the given point (xp,yp) is located within the field of view
    of the ZTF camera.
    (xp,yp) are the coordinates of the given object in the ortographic
    projection with the origin at the center of the field of view.
    This code checks if the point is located within any of ZTF CCDs,
    if yes, it returns the ccd and quadrant number. If not, None is
    returned.
    (x,y) are arrays containing the coordinates of CCD vertices, see
    http://www.oir.caltech.edu/twiki_ptf/pub/ZTF/ZTFFieldGrid/ZTF_CCD_Layout.tbl

    Parameters
    ----------
    xp : float
        x-coordinate of the object
    yp : float
        y-coordinate of the object
    x : list
        x-coordinates of CCD vertices
    y : list
        y-coordinates of CCD vertices

    Returns
    -------
    tuple
        ccd, quadrant number
    """

    ccd_quads = []

    for i in range(16):
        idx = 4 * i

        y_test_1 = fit_line(
            xp, x[idx], y[idx] - CCD_BUFFER, x[idx + 3], y[idx + 3] - CCD_BUFFER
        )
        y_test_2 = fit_line(
            xp, x[idx + 1], y[idx + 1] + CCD_BUFFER, x[idx + 2], y[idx + 2] + CCD_BUFFER
        )

        x_test_1 = fit_line(
            yp, y[idx], x[idx] - CCD_BUFFER, y[idx + 1], x[idx + 1] - CCD_BUFFER
        )
        x_test_2 = fit_line(
            yp, y[idx + 3], x[idx + 3] + CCD_BUFFER, y[idx + 2], x[idx + 2] + CCD_BUFFER
        )

        if yp < y_test_1 or yp > y_test_2:
            continue
        if xp < x_test_1 or xp > x_test_2:
            continue

        ccd = i + 1

        y_test = fit_line(
            xp,
            0.5 * (x[idx + 2] + x[idx + 3]),
            0.5 * (y[idx + 2] + y[idx + 3]),
            0.5 * (x[idx] + x[idx + 1]),
            0.5 * (y[idx] + y[idx + 1]),
        )
        x_test = fit_line(
            yp,
            0.5 * (y[idx] + y[idx + 3]),
            0.5 * (x[idx] + x[idx + 3]),
            0.5 * (y[idx + 1] + y[idx + 2]),
            0.5 * (x[idx + 1] + x[idx + 2]),
        )

        # also use a buffer here, and a point could be in multiple quadrants
        if yp < y_test or abs(yp - y_test) < CCD_BUFFER:
            if xp < x_test or abs(xp - x_test) < CCD_BUFFER:
                ccd_quads.append((ccd, 4))
            if xp > x_test or abs(xp - x_test) < CCD_BUFFER:
                ccd_quads.append((ccd, 3))

        if yp > y_test or abs(yp - y_test) < CCD_BUFFER:
            if xp < x_test or abs(xp - x_test) < CCD_BUFFER:
                ccd_quads.append((ccd, 1))
            if xp > x_test or abs(xp - x_test) < CCD_BUFFER:
                ccd_quads.append((ccd, 2))

    return ccd_quads


@njit
def great_circle_distance_rad(
    ra1: float, dec1: float, ra2: float, dec2: float
) -> float:
    """
    Distance between two points on the sphere

    Parameters
    ----------
    ra1 : float
        Right ascension of the first object
    dec1 : float
        Declination of the first object
    ra2 : float
        Right ascension of the second object
    dec2 : float
        Declination of the second object

    Returns
    -------
    float
        Distance between two points in radians
    """
    delta_ra = np.abs(ra2 - ra1)
    distance = np.arctan2(
        np.sqrt(
            (np.cos(dec2) * np.sin(delta_ra)) ** 2
            + (
                np.cos(dec1) * np.sin(dec2)
                - np.sin(dec1) * np.cos(dec2) * np.cos(delta_ra)
            )
            ** 2
        ),
        np.sin(dec1) * np.sin(dec2) + np.cos(dec1) * np.cos(dec2) * np.cos(delta_ra),
    )
    return distance


@njit
def great_circle_distance(
    ra1_deg: float, dec1_deg: float, ra2_deg: float, dec2_deg: float
) -> float:
    """
    Distance between two points on the sphere

    Parameters
    ----------
    ra1_deg : float
        Right ascension of the first object in degrees
    dec1_deg : float
        Declination of the first object in degrees
    ra2_deg : float
        Right ascension of the second object in degrees
    dec2_deg : float
        Declination of the second object in degrees

    Returns
    -------
    float
        Distance between two points in degrees
    """
    ra1, dec1, ra2, dec2 = (
        ra1_deg * DEGRA,
        dec1_deg * DEGRA,
        ra2_deg * DEGRA,
        dec2_deg * DEGRA,
    )
    return great_circle_distance_rad(ra1, dec1, ra2, dec2) / DEGRA


def get_field_id(ra: float, dec: float, radius: Union[float, int, None] = None) -> list:
    """
    Find the field, ccd, and quadrant number for a given object.
    ra, dec are coordinates of the object (in degrees), fieldno,
    ra_all, and dec_all are arrays containing ZTF field numbers,
    and coordinates of their centers (ra_all,dec_all).
    These calculations are approximate and may fail if the object
    is located close to the edge of the field of view / edge of the
    reference image.

    Parameters
    ----------
    ra : float
        Right ascension of the object in degrees
    dec : float
        Declination of the object in degrees
    radius : Union[float, int, None], optional
        Radius of the search in arcsec, by default None.

    Returns
    -------
    list
        List of field, ccd, and quadrant number
    """

    ra = ra * DEGRA
    dec = dec * DEGRA

    field_adist = ADIST_MAX
    if radius is not None:
        field_adist += (radius / 3600) * DEGRA

    nf = len(ZTFFieldData.fieldno)

    res = []

    for i in range(nf):
        adist = great_circle_distance_rad(
            ra, dec, ZTFFieldData.ra_all[i], ZTFFieldData.dec_all[i]
        )
        if adist > field_adist:
            continue
        x, y = ortographic_projection(
            ra, dec, ZTFFieldData.ra_all[i], ZTFFieldData.dec_all[i]
        )
        # TODO, use the radius (if provided) to increase the CCD buffer size
        ccd_quads = inside_polygon(x, y, CCD_LAYOUT_X, CCD_LAYOUT_Y)
        if len(ccd_quads) > 0:
            for ccd, quad in ccd_quads:
                res.append([ZTFFieldData.fieldno[i], ccd, quad])

    return res



import argparse
import os

from zvartools.enums import ALLOWED_BANDS


def candidates_parser() -> argparse.ArgumentParser:
    """
    Create the parser for the candidate catalog creation script

    Returns
    -------
    argparse.ArgumentParser
        Parser for the candidate catalog creation script
    """
    parser = argparse.ArgumentParser(
        description="Run candidate detection on ZVAR results"
    )
    parser.add_argument(
        "--field_min",
        type=int,
        help="Field number on the low end of the range (inclusive)",
    )
    parser.add_argument(
        "--field_max",
        type=int,
        help="Field number on the high end of the range (inclusive)",
    )
    parser.add_argument(
        "--bands", type=str, default="g,r", help="Bands to process, separated by commas"
    )
    parser.add_argument(
        "--radius",
        type=float,
        default=2.0,
        help="Radius in arcseconds to search for xmatches in external catalogs",
    )
    parser.add_argument(
        "--periods_path",
        default="./data/periods",
        type=str,
        help="Path where to load the ZVAR files (e.g. periods for a field)",
    )
    parser.add_argument(
        "--output_path", type=str, help="Output directory for the CSV files"
    )
    parser.add_argument(
        "--credentials_path",
        type=str,
        help="Path to the JSON file with the Kowalski credentials",
    )

    return parser


def validate_candidates_args(args) -> argparse.Namespace:
    """
    Validate the arguments for the candidate catalog creation script

    Parameters
    ----------
    args : argparse.Namespace
        Arguments parsed by the parser

    Returns
    -------
    argparse.Namespace
        Validated arguments

    Raises
    ------
    ValueError
        If any of the arguments is invalid
    """
    # FIELD_MIN
    try:
        args.field_min = int(args.field_min)
    except ValueError:
        raise ValueError("Field min must be an integer")

    # FIELD_MAX
    try:
        args.field_max = int(args.field_max)
    except ValueError:
        raise ValueError("Field max must be an integer")

    # BANDS
    args.bands = [band.strip() for band in args.bands.split(",")]
    if not all(band in ALLOWED_BANDS for band in args.bands):
        raise ValueError(f"Invalid band specified. Allowed bands are {ALLOWED_BANDS}")

    # RADIUS
    try:
        args.radius = float(args.radius)
    except ValueError:
        raise ValueError("Radius must be a number (integer or float, in arcseconds)")

    if args.radius < 0.0 or args.radius > 180.0:
        raise ValueError("Radius must be between 0 and 3600 arcseconds")

    # PERIODS_PATH
    if not os.path.exists(args.periods_path):
        raise ValueError(f"Periods path {args.periods_path} does not exist")

    # OUTPUT_PATH
    if not os.path.exists(args.output_path):
        raise ValueError(f"Output path {args.output_path} does not exist")

    # CREDENTIALS
    if args.credentials_path is None:
        raise ValueError("Credentials file must be specified")
    if not os.path.exists(args.credentials_path):
        raise ValueError(f"Credentials file {args.credentials_path} does not exist")

    return args



import os
from typing import List, Union

import numpy as np
import pandas as pd

from zvartools.external import (
    Kowalski,
    query_gaia,
    query_ps1,
    query_2mass,
    query_allwise,
)
from zvartools.stats import get_cdf, get_extrema
from zvartools.enums import BIN_IDX_TO_FREQ_COL, BIN_IDX_TO_FAP_COL


class PS1Match:
    def __init__(self, id, g, g_err, r, r_err, i, i_err, z, z_err, y, y_err):
        self.id = id
        self.g = g
        self.g_err = g_err
        self.r = r
        self.r_err = r_err
        self.i = i
        self.i_err = i_err
        self.z = z
        self.z_err = z_err
        self.y = y
        self.y_err = y_err

    def __repr__(self):
        return f"G: {self.g:.2f}±{self.g_err:.2f}, R: {self.r:.2f}±{self.r_err:.2f}, I: {self.i:.2f}±{self.i_err:.2f}, Z: {self.z:.2f}±{self.z_err:.2f}, Y: {self.y:.2f}±{self.y_err:.2f}"

    def __str__(self):
        return self.__repr__()


class GaiaMatch:
    def __init__(
        self,
        id,
        G,
        BP,
        RP,
        parallax,
        parallax_error,
        pmra,
        pmra_error,
        pmdec,
        pmdec_error,
        BP_RP=None,
        MG=None,
    ):
        self.id = id
        self.G = G
        self.BP = BP
        self.RP = RP
        self.parallax = parallax
        self.parallax_error = parallax_error
        self.pmra = pmra
        self.pmra_error = pmra_error
        self.pmdec = pmdec
        self.pmdec_error = pmdec_error

        self.add_bp_rp()
        self.add_MG()

    def __repr__(self):
        return f"G: {self.G:.2f}, BP: {self.BP:.2f}, RP: {self.RP:.2f}, Parallax: {self.parallax:.2f}±{self.parallax_error:.2f}, PMRA: {self.pmra:.2f}±{self.pmra_error:.2f}, PMDEC: {self.pmdec:.2f}±{self.pmdec_error:.2f}"

    def __str__(self):
        return self.__repr__()

    def add_bp_rp(self):
        if self.BP is not None and self.RP is not None:
            self.BP_RP = self.BP - self.RP

    def add_MG(self):
        if self.G is not None and self.parallax is not None:
            if (
                self.parallax / 1000 <= 0
                or np.isnan(self.parallax)
                or self.parallax is None
            ):
                self.MG = None
            else:
                self.MG = self.G + 5 * np.log10(self.parallax / 1000) + 5


class TwoMASSMatch:
    def __init__(self, id, j, j_err, h, h_err, k, k_err):
        self.id = id
        self.j = j
        self.j_err = j_err
        self.h = h
        self.h_err = h_err
        self.k = k
        self.k_err = k_err

    def __repr__(self):
        return f"J: {self.j:.2f}±{self.j_err:.2f}, H: {self.h:.2f}±{self.h_err:.2f}, K: {self.k:.2f}±{self.k_err:.2f}"

    def __str__(self):
        return self.__repr__()


class AllWISEMatch:
    def __init__(self, id, w1, w1_err, w2, w2_err, w3, w3_err, w4, w4_err):
        self.id = id
        self.w1 = w1
        self.w1_err = w1_err
        self.w2 = w2
        self.w2_err = w2_err
        self.w3 = w3
        self.w3_err = w3_err
        self.w4 = w4
        self.w4_err = w4_err

    def __repr__(self):
        return f"W1: {self.w1:.2f}±{self.w1_err:.2f}, W2: {self.w2:.2f}±{self.w2_err:.2f}, W3: {self.w3:.2f}±{self.w3_err:.2f}, W4: {self.w4:.2f}±{self.w4_err:.2f}"

    def __str__(self):
        return self.__repr__()


class VariabilityCandidate:
    def __init__(
        self,
        psid,
        ra,
        dec,
        valid,
        freq,
        fap,
        best_M,
        field: int = None,
        ccd: int = None,
        quad: int = None,
        band: str = None,
        ps1: Union[PS1Match, dict] = None,
        gaia: Union[GaiaMatch, dict] = None,
        twomass: Union[TwoMASSMatch, dict] = None,
        allwise: Union[AllWISEMatch, dict] = None,
    ):
        self.psid = psid
        self.ra = ra
        self.dec = dec
        self.valid = valid
        self.freq = freq
        self.fap = fap
        self.best_M = best_M
        self.field = field
        self.ccd = ccd
        self.quad = quad
        self.band = band
        self.set_ps1(ps1)
        self.set_gaia(gaia)
        self.set_2mass(twomass)
        self.set_allwise(allwise)

    @property
    def period(self):
        """
        Calculate the period of the candidate, in hours
        """
        if self.freq is None:
            return None
        return 1 / self.freq * 24

    def set_ps1(
        self,
        ps1: Union[PS1Match, dict, None],
    ):
        """
        Set the PS1 data for the candidate

        Parameters
        ----------
        ps1 : Union[PS1Match, dict, None]
            PS1 data

        Raises
        ------
        ValueError
            If the PS1 data is not of the correct type
        """
        if isinstance(ps1, PS1Match):
            self.ps1 = ps1
        elif isinstance(ps1, dict):
            self.ps1 = PS1Match(**ps1)
        elif ps1 is None:
            self.ps1 = None
        else:
            raise ValueError("PS1 must be a PS1Match, a dictionary, or None")

    def set_gaia(
        self,
        gaia: Union[GaiaMatch, dict, None],
    ):
        """
        Set the Gaia data for the candidate

        Parameters
        ----------
        gaia : Union[GaiaMatch, dict, None]
            Gaia data

        Raises
        ------
        ValueError
            If the Gaia data is not of the correct type
        """
        if isinstance(gaia, GaiaMatch):
            self.gaia = gaia
        elif isinstance(gaia, dict):
            self.gaia = GaiaMatch(**gaia)
        elif gaia is None:
            self.gaia = None
        else:
            raise ValueError("Gaia must be a GaiaMatch, a dictionary, or None")

    def set_2mass(
        self,
        twomass: Union[TwoMASSMatch, dict, None],
    ):
        """
        Set the 2MASS data for the candidate

        Parameters
        ----------
        twomass : Union[TwoMASSMatch, dict, None]
            2MASS data

        Raises
        ------
        ValueError
            If the 2MASS data is not of the correct type
        """
        if isinstance(twomass, TwoMASSMatch):
            self.twomass = twomass
        elif isinstance(twomass, dict):
            self.twomass = TwoMASSMatch(**twomass)
        elif twomass is None:
            self.twomass = None
        else:
            raise ValueError("2MASS must be a TwoMASSMatch, a dictionary, or None")

    def set_allwise(
        self,
        allwise: Union[AllWISEMatch, dict, None],
    ):
        """
        Set the AllWISE data for the candidate

        Parameters
        ----------
        allwise : Union[AllWISEMatch, dict, None]
            AllWISE data

        Raises
        ------
        ValueError
            If the AllWISE data is not of the correct type
        """
        if isinstance(allwise, AllWISEMatch):
            self.allwise = allwise
        elif isinstance(allwise, dict):
            self.allwise = AllWISEMatch(**allwise)
        elif allwise is None:
            self.allwise = None
        else:
            raise ValueError("AllWISE must be a AllWISEMatch, a dictionary, or None")

    def __repr__(self):
        return (
            f"ID: {self.psid}, RA: {self.ra}, Dec: {self.dec}, Valid: {self.valid}, Best M: {self.best_M}, freq: {self.freq}, FAP: {self.fap}"
            + (
                f" (field: {self.field}, ccd: {self.ccd}, quad: {self.quad})"
                if self.field and self.ccd and self.quad
                else ""
            )
            + (f"\n    PS1: {self.ps1}" if self.ps1 else "")
            + (f"\n    Gaia: {self.gaia}" if self.gaia else "")
            + (f"\n    2MASS: {self.twomass}" if self.twomass else "")
            + (f"\n    AllWISE: {self.allwise}" if self.allwise else "")
        )


def get_candidates(
    psids: np.ndarray,
    ra: np.ndarray,
    dec: np.ndarray,
    ratio_valid: np.ndarray,
    freqs: np.ndarray,
    sigs: np.ndarray,
    field: int,
    ccds: np.ndarray,
    quads: np.ndarray,
) -> List[VariabilityCandidate]:
    """
    Get the variability candidates from the data

    Parameters
    ----------
    psids : np.ndarray
        PS1 IDs
    ra : np.ndarray
        Right ascensions
    dec : np.ndarray
        Declinations
    ratio_valid : np.ndarray
        Ratio of valid data points
    freqs : np.ndarray
        Frequencies
    sigs : np.ndarray
        Significances
    field : int
        Field number
    ccds : np.ndarray
        CCD numbers
    quads : np.ndarray
        Quadrant numbers

    Returns
    -------
    List[VariabilityCandidate]
        List of variability candidates
    """
    # Precompute the CDFs for each bin
    cdfs = [get_cdf(sigs[:, j, 0]) for j in range(3)]

    extrema_20 = get_extrema(sigs[:, 0, 0], 0.999)
    print(f"Found {np.sum(extrema_20)} candidates in the 20-day bin")
    extrema_10 = get_extrema(sigs[:, 1, 0], 0.999)
    print(f"Found {np.sum(extrema_10)} candidates in the 10-day bin")
    extrema_5 = get_extrema(sigs[:, 2, 0], 0.999)
    print(f"Found {np.sum(extrema_5)} candidates in the 5-day bin")
    combined_candidates = np.logical_or.reduce([extrema_20, extrema_10, extrema_5])
    print(f"Found {np.sum(combined_candidates)} combined candidates")

    candidate_list = []
    bins = [20, 10, 5]
    for i in range(len(psids)):
        if combined_candidates[i]:
            # Find the probability of the 0th value in each bin
            probabilities = []
            for j in range(3):
                bin_centers, cdf = cdfs[j]
                value = sigs[i, j, 0]
                if value > 0:
                    prob_index = np.searchsorted(bin_centers, value)
                    prob = cdf[prob_index] if prob_index < len(cdf) else 1.0
                else:
                    prob = 1.0  # Assign a high probability for non-positive values
                probabilities.append(1 - prob)

            # Determine which bin value is the least likely to have arisen from random chance
            min_prob_index = np.argmin(probabilities)
            # min_prob_value = probabilities[min_prob_index]
            # Determine the best bin
            best_M = bins[min_prob_index]

            # Append the candidate with probabilities instead of significance values
            candidate_list.append(
                VariabilityCandidate(
                    psids[i],
                    ra[i],
                    dec[i],
                    ratio_valid[i],
                    freqs[i, :, 0],
                    probabilities,
                    best_M,
                    field,
                    ccds[i],
                    quads[i],
                )
            )

    return candidate_list


def add_ps1_xmatch_to_candidates(
    k: Kowalski, candidate_list: List[VariabilityCandidate]
) -> List[VariabilityCandidate]:
    """
    Add PS1 crossmatches to the candidates

    Parameters
    ----------
    k : Kowalski
        Kowalski instance
    candidate_list : List[VariabilityCandidate]
        List of variability candidates

    Returns
    -------
    List[VariabilityCandidate]
        List of variability candidates with PS1 crossmatches
    """
    # query_ps1 only needs the psids
    psids = [candidate.psid for candidate in candidate_list]
    xmatches = query_ps1(k, psids)
    # print how many sources have ps1 matches
    print(
        f"Found {len([x for x in xmatches.values() if x is not None])} PS1 matches, out of {len(candidate_list)} sources"
    )
    # Fill in the PS1 data for each candidate
    for candidate in candidate_list:
        result = xmatches.get(candidate.psid)
        if result:
            candidate.set_ps1(
                PS1Match(
                    candidate.psid,
                    result.get("gMeanPSFMag"),
                    result.get("gMeanPSFMagErr"),
                    result.get("rMeanPSFMag"),
                    result.get("rMeanPSFMagErr"),
                    result.get("iMeanPSFMag"),
                    result.get("iMeanPSFMagErr"),
                    result.get("zMeanPSFMag"),
                    result.get("zMeanPSFMagErr"),
                    result.get("yMeanPSFMag"),
                    result.get("yMeanPSFMagErr"),
                )
            )

    return candidate_list


def add_gaia_xmatch_to_candidates(
    k: Kowalski, candidate_list: List[VariabilityCandidate], radius: float = 2.0
) -> List[VariabilityCandidate]:
    """
    Add Gaia crossmatches to the candidates

    Parameters
    ----------
    k : Kowalski
        Kowalski instance
    candidate_list : List[VariabilityCandidate]
        List of variability candidates
    radius : float
        Radius in arcseconds to search for xmatches in external catalogs

    Returns
    -------
    List[VariabilityCandidate]
        List of variability candidates with Gaia crossmatches
    """
    if not isinstance(candidate_list, list):
        raise ValueError("Candidates must be provided as a list")
    if not all(
        isinstance(candidate, VariabilityCandidate) for candidate in candidate_list
    ):
        raise ValueError("Candidates must be of type Candidate")

    # Extract the psids, ras, and decs from the candidate list
    psids = [candidate.psid for candidate in candidate_list]
    ras = [candidate.ra for candidate in candidate_list]
    decs = [candidate.dec for candidate in candidate_list]

    xmatches = query_gaia(k, psids, ras, decs, radius)
    # print how many sources have gaia matches
    print(
        f"Found {len([x for x in xmatches.values() if x is not None])} Gaia matches, out of {len(candidate_list)} sources"
    )
    # Fill in the Gaia data for each candidate
    for candidate in candidate_list:
        result = xmatches.get(candidate.psid)
        if result and result.get("id"):  # If Gaia found anything
            candidate.set_gaia(
                GaiaMatch(
                    result.get("id"),
                    result.get("phot_g_mean_mag"),
                    result.get("phot_bp_mean_mag"),
                    result.get("phot_rp_mean_mag"),
                    result.get("parallax"),
                    result.get("parallax_error"),
                    result.get("pmra"),
                    result.get("pmra_error"),
                    result.get("pmdec"),
                    result.get("pmdec_error"),
                )
            )

    return candidate_list


def add_2mass_xmatch_to_candidates(
    k: Kowalski, candidate_list: List[VariabilityCandidate], radius: float = 2.0
) -> List[VariabilityCandidate]:
    """
    Add 2MASS crossmatches to the candidates

    Parameters
    ----------
    k : Kowalski
        Kowalski instance
    candidate_list : List[VariabilityCandidate]
        List of variability candidates
    radius : float
        Radius in arcseconds to search for xmatches in external catalogs

    Returns
    -------
    List[VariabilityCandidate]
        List of variability candidates with 2MASS crossmatches
    """
    if not isinstance(candidate_list, list):
        raise ValueError("Candidates must be provided as a list")
    if not all(
        isinstance(candidate, VariabilityCandidate) for candidate in candidate_list
    ):
        raise ValueError("Candidates must be of type Candidate")

    # Extract the psids, ras, and decs from the candidate list
    psids = [candidate.psid for candidate in candidate_list]
    ras = [candidate.ra for candidate in candidate_list]
    decs = [candidate.dec for candidate in candidate_list]

    xmatches = query_2mass(k, psids, ras, decs, radius)
    # print how many sources have 2mass matches
    print(
        f"Found {len([x for x in xmatches.values() if x is not None])} 2MASS matches, out of {len(candidate_list)} sources"
    )
    # Fill in the 2MASS data for each candidate
    for candidate in candidate_list:
        result = xmatches.get(candidate.psid)
        if result and result.get("id"):  # If 2MASS found anything
            candidate.set_2mass(
                TwoMASSMatch(
                    result.get("id"),
                    result.get("j_m"),
                    result.get("j_msigcom"),
                    result.get("h_m"),
                    result.get("h_msigcom"),
                    result.get("k_m"),
                    result.get("k_msigcom"),
                )
            )

    return candidate_list


def add_allwise_xmatch_to_candidates(
    k: Kowalski, candidate_list: List[VariabilityCandidate], radius: float = 2.0
) -> List[VariabilityCandidate]:
    """
    Add AllWISE crossmatches to the candidates

    Parameters
    ----------
    k : Kowalski
        Kowalski instance
    candidate_list : List[VariabilityCandidate]
        List of variability candidates
    radius : float
        Radius in arcseconds to search for xmatches in external catalogs

    Returns
    -------
    List[VariabilityCandidate]
        List of variability candidates with AllWISE crossmatches
    """
    if not isinstance(candidate_list, list):
        raise ValueError("Candidates must be provided as a list")
    if not all(
        isinstance(candidate, VariabilityCandidate) for candidate in candidate_list
    ):
        raise ValueError("Candidates must be of type Candidate")

    # Extract the psids, ras, and decs from the candidate list
    psids = [candidate.psid for candidate in candidate_list]
    ras = [candidate.ra for candidate in candidate_list]
    decs = [candidate.dec for candidate in candidate_list]

    xmatches = query_allwise(k, psids, ras, decs, radius)
    # print how many sources have allwise matches
    print(
        f"Found {len([x for x in xmatches.values() if x is not None])} AllWISE matches, out of {len(candidate_list)} sources"
    )
    # Fill in the AllWISE data for each candidate
    for candidate in candidate_list:
        result = xmatches.get(candidate.psid)
        if result and result.get("id"):  # If AllWISE found anything
            candidate.set_allwise(
                AllWISEMatch(
                    result.get("id"),
                    result.get("w1mpro"),
                    result.get("w1sigmpro"),
                    result.get("w2mpro"),
                    result.get("w2sigmpro"),
                    result.get("w3mpro"),
                    result.get("w3sigmpro"),
                    result.get("w4mpro"),
                    result.get("w4sigmpro"),
                )
            )

    return candidate_list


def export_to_parquet(
    candidate_list: List[VariabilityCandidate], field: int, band: int, path: str
):
    """
    Export the candidates to a parquet file

    Parameters
    ----------
    candidate_list : List[VariabilityCandidate]
        List of variability candidates
    field : int
        Field number
    band : int
        Band number
    path : str
        Output directory

    Raises
    ------
    ValueError
        If any of the arguments is invalid
    """
    if not isinstance(candidate_list, list):
        raise ValueError("Candidates must be provided as a list")
    if not all(
        isinstance(candidate, VariabilityCandidate) for candidate in candidate_list
    ):
        raise ValueError("Candidates must be of type Candidate")
    if not isinstance(field, (int, np.integer, str)):
        raise ValueError("Field must be an integer or string")
    if not isinstance(band, (int, np.integer, str)):
        raise ValueError("Band must be an integer or string")
    if path is None:
        raise ValueError("Output directory must be specified")

    # before we export the data, we want to sort the candidates by their id (psid)
    # to make sure we get consistent results when rerunning the code
    candidate_list = sorted(candidate_list, key=lambda x: x.psid)

    data = []
    for candidate in candidate_list:
        candidate_data = {
            "psid": candidate.psid,
            "ra": candidate.ra,
            "dec": candidate.dec,
            "valid": candidate.valid,
            "best_M": candidate.best_M,
            "field": candidate.field,
            "ccd": candidate.ccd,
            "quad": candidate.quad,
            "frequency_20": candidate.freq[0],
            "frequency_10": candidate.freq[1],
            "frequency_5": candidate.freq[2],
            "FAP_20": candidate.fap[0],
            "FAP_10": candidate.fap[1],
            "FAP_5": candidate.fap[2],
        }
        if candidate.ps1:
            candidate_data.update(
                {
                    f"ps_{key}": getattr(candidate.ps1, key)
                    for key in vars(candidate.ps1)
                    if key != "id"
                }
            )
        if candidate.gaia:
            candidate_data.update(
                {
                    f"gaia_{key}": getattr(candidate.gaia, key)
                    for key in vars(candidate.gaia)
                }
            )
        if candidate.twomass:
            candidate_data.update(
                {
                    f"twomass_{key}": getattr(candidate.twomass, key)
                    for key in vars(candidate.twomass)
                }
            )
        if candidate.allwise:
            candidate_data.update(
                {
                    f"allwise_{key}": getattr(candidate.allwise, key)
                    for key in vars(candidate.allwise)
                }
            )
        data.append(candidate_data)
    # Construct the candidate path
    candidate_path = os.path.join(
        os.path.abspath(path),
        f"{str(field).zfill(4)}",
        f"fpw_{str(field).zfill(4)}_z{band}.parquet",
    )

    # Check if the directory exists, and create it if it doesn't
    os.makedirs(os.path.dirname(candidate_path), exist_ok=True)

    # Create the DataFrame and write to parquet
    pd.DataFrame(data).to_parquet(candidate_path, index=False)


def import_from_parquet(path: str, best_m_only=True) -> List[VariabilityCandidate]:
    """
    Import the candidates from a parquet file

    Parameters
    ----------
    path : str
        Path to the parquet file
    best_m_only : bool
        Whether to only use the best M value

    Returns
    -------
    List[VariabilityCandidate]
        List of variability candidates
    """
    if not os.path.exists(path):
        raise ValueError(f"File {path} does not exist")
    df = pd.read_parquet(path)
    candidate_list = []

    # get the filter from the path name
    try:
        band = str(path.split("_z")[-1].split(".")[0])
    except ValueError:
        band = None

    for _, row in df.iterrows():
        ps1_data = (
            {
                key.replace("ps_", ""): row[key]
                for key in row.keys()
                if key.startswith("ps_")
            }
            if any(key.startswith("ps_") for key in row.keys())
            else None
        )
        ps1_data.update({"id": row["psid"]}) if ps1_data else None

        gaia_data = (
            {
                key.replace("gaia_", ""): row[key]
                for key in row.keys()
                if key.startswith("gaia_")
            }
            if row.get("gaia_id")
            else None
        )

        twomass_data = (
            {
                key.replace("twomass_", ""): row[key]
                for key in row.keys()
                if key.startswith("twomass_")
            }
            if row.get("twomass_id")
            else None
        )

        allwise_data = (
            {
                key.replace("allwise_", ""): row[key]
                for key in row.keys()
                if key.startswith("allwise_")
            }
            if row.get("allwise_id")
            else None
        )

        if not best_m_only:
            candidate = VariabilityCandidate(
                psid=row["psid"],
                ra=row["ra"],
                dec=row["dec"],
                valid=row["valid"],
                freq=np.array(
                    [row["frequency_20"], row["frequency_10"], row["frequency_5"]]
                ),
                fap=np.array([row["FAP_20"], row["FAP_10"], row["FAP_5"]]),
                best_M=row["best_M"],
                field=row.get("field"),
                ccd=row.get("ccd"),
                quad=row.get("quad"),
                band=band,
                ps1=ps1_data,
                gaia=gaia_data,
                twomass=twomass_data,
                allwise=allwise_data,
            )
        else:
            bin_idx = 0
            if row["best_M"] == 10:
                bin_idx = 1
            elif row["best_M"] == 5:
                bin_idx = 2

            candidate = VariabilityCandidate(
                psid=row["psid"],
                ra=row["ra"],
                dec=row["dec"],
                valid=row["valid"],
                freq=row[BIN_IDX_TO_FREQ_COL[bin_idx]],
                fap=row[BIN_IDX_TO_FAP_COL[bin_idx]],
                best_M=row["best_M"],
                field=row.get("field"),
                ccd=row.get("ccd"),
                quad=row.get("quad"),
                band=band,
                ps1=ps1_data,
                gaia=gaia_data,
                twomass=twomass_data,
                allwise=allwise_data,
            )
        candidate_list.append(candidate)

    return candidate_list



from zvartools.spatial import great_circle_distance

import numpy as np
import time

# generate 12k random points on the sphere
n = 12000
ra1 = np.random.uniform(0, 360, n)
dec1 = np.random.uniform(-90, 90, n)
ra2 = np.random.uniform(0, 360, n)
dec2 = np.random.uniform(-90, 90, n)

# time how long it takes to calculate the distances
start = time.time()
for i in range(n):
    great_circle_distance(ra1[i], dec1[i], ra2[i], dec2[i])
end = time.time()

print(f"Great circle distance calculation for {n} pairs took {end - start:.2f} seconds")
